from PySide6.QtWidgets import QHBoxLayout, QPushButton, QLabel, QSizePolicy, QTextEdit, QVBoxLayout, QWidget, QStyledItemDelegate, QStyle, QListView, QLineEdit, QFrame
from PySide6.QtGui import QFontMetrics, QFont, QColor
from PySide6.QtCore import Qt, QSize, QAbstractListModel, QRect, QMargins

from subtitld.interface import subtitles_panel
from subtitld.modules import utils
from subtitld.modules import quality_check
from subtitld.modules import subtitles


class subtitles_panel_qlistwidget_model(QAbstractListModel):
    def __init__(self, *args, subtitles=None, **kwargs):
        super(subtitles_panel_qlistwidget_model, self).__init__(*args, **kwargs)
        self.subtitles = subtitles or []

    def data(self, index, role):
        if role == Qt.DisplayRole:
            return self.subtitles[index.row()][-1]

        # if role == Qt.DecorationRole:
        #     status, _ = self.subtitles[index.row()]
        #     if status:
        #         return tick

    def get_index(self, subtitle):
        index = self.subtitles.index(subtitle)
        return self.index(index)

    def rowCount(self, index):
        return len(self.subtitles)


class subtitles_panel_qlistwidget_delegate(QStyledItemDelegate):
    def __init__(self, parent=None, settings=None):
        super(subtitles_panel_qlistwidget_delegate, self).__init__(parent)
        self.settings = settings or {}

    def get_number_width(self, index):
        number_width = QFontMetrics(QFont('Ubuntu', 8)).horizontalAdvance((len(str(index.model().rowCount(index)))) * '8')
        return number_width

    def get_text_height(self, option, index):
        row_text = index.data(Qt.DisplayRole)
        width = option.rect.width()
        height = QFontMetrics(QFont('Ubuntu', 11)).boundingRect(QRect(0, 0, width - (20 + 10 + self.get_number_width(index) + 10 + 10), 100), Qt.TextWordWrap, row_text).height()
        return height

    def paint(self, painter, option, index):
        row_text = index.data(Qt.DisplayRole)
        number_width = self.get_number_width(index)

        if option.state & QStyle.State_Selected:
            painter.setPen(Qt.NoPen)
            painter.setBrush(QColor(self.settings.get('subtitle_list', {}).get('background_color', '#102e3e4c')))
            painter.drawRect(option.rect)
            # painter.setBrush(QColor('#55d43f'))
            # painter.drawRect(QRect(0, option.rect.y(), 20, option.rect.height()))

        sub_is_ok = True
        if self.settings['quality_check'].get('enabled', False):
            sub_is_ok, _, _ = quality_check.check_subtitle(index.model().subtitles[index.row()], self.settings['quality_check'])

        painter.setPen(Qt.NoPen)
        painter.setBrush(QColor(self.settings.get('subtitle_list', {}).get('number_background_color', '#aa2e3e4c') if sub_is_ok else self.settings.get('subtitle_list', {}).get('number_background_warning_color', '#aa9e1a1a')))

        number_rect = QRect(20, option.rect.y(), 10 + number_width + 10, option.rect.height())
        text_rect = QRect(number_rect.right(), option.rect.y(), option.rect.width() - number_rect.right(), option.rect.height())

        painter.drawRect(number_rect)

        painter.setPen(QColor(self.settings.get('subtitle_list', {}).get('number_color', '#ffffff') if sub_is_ok else self.settings.get('subtitle_list', {}).get('number_warning_color', '#ffffff')))

        number_rect = number_rect.marginsRemoved(QMargins(10, 10, 10, 10))

        painter.setFont(QFont('Ubuntu', 8))
        painter.drawText(number_rect, 0, str(index.row() + 1))

        painter.setPen(Qt.NoPen)
        painter.setBrush(QColor(self.settings.get('subtitle_list', {}).get('text_background_color', '#102e3e4c') if sub_is_ok else self.settings.get('subtitle_list', {}).get('text_background_warning_color', '#109e1a1a')))

        painter.drawRect(text_rect)

        text_rect = text_rect.marginsRemoved(QMargins(10, 10, 10, 10))

        painter.setPen(QColor(self.settings.get('subtitle_list', {}).get('text_color', '#2e3e4c') if sub_is_ok else self.settings.get('subtitle_list', {}).get('text_warning_color', '#9e1a1a')))
        painter.setFont(QFont('Ubuntu', 11))
        painter.drawText(text_rect, Qt.TextWordWrap, row_text)

    def sizeHint(self, option, index):
        width = option.rect.width()
        height = self.get_text_height(option, index) + 20
        return QSize(width, height)


def add_widgets(self):
    self.subtitles_panel_simplelist_widget = QWidget()
    self.subtitles_panel_simplelist_widget_vbox = QVBoxLayout(self.subtitles_panel_simplelist_widget)
    self.subtitles_panel_simplelist_widget_vbox.setContentsMargins(0, 0, 0, 0)
    self.subtitles_panel_simplelist_widget_vbox.setSpacing(10)

    self.subtitles_panel_qlistwidget_model = subtitles_panel_qlistwidget_model()

    self.subtitles_panel_qlistwidget_delegate = subtitles_panel_qlistwidget_delegate(settings=self.settings)

    self.subtitles_panel_qlistwidget = QListView()
    self.subtitles_panel_qlistwidget.setViewMode(QListView.ListMode)
    self.subtitles_panel_qlistwidget.setObjectName('subtitles_panel_qlistwidget')
    self.subtitles_panel_qlistwidget.setContentsMargins(QMargins(0, 0, 0, 0))
    self.subtitles_panel_qlistwidget.setSpacing(0)
    self.subtitles_panel_qlistwidget.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
    self.subtitles_panel_qlistwidget.setFocusPolicy(Qt.NoFocus)
    self.subtitles_panel_qlistwidget.setModel(self.subtitles_panel_qlistwidget_model)
    self.subtitles_panel_qlistwidget.setItemDelegate(self.subtitles_panel_qlistwidget_delegate)
    # self.subtitles_panel_qlistwidget.setSizePolicy(QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Maximum))
    self.subtitles_panel_qlistwidget.clicked.connect(lambda: subtitles_panel_qlistwidget_item_clicked(self))
    self.subtitles_panel_simplelist_widget_vbox.addWidget(self.subtitles_panel_qlistwidget, 1)

    self.properties_information = QFrame()
    self.properties_information.setObjectName('properties_information')
    self.properties_information.setLayout(QVBoxLayout())
    self.properties_information.layout().setContentsMargins(20, 0, 0, 0)
    self.properties_information.layout().setSpacing(5)

    self.properties_information_stats = QFrame()
    self.properties_information_stats.setObjectName('properties_information')
    self.properties_information_stats.setLayout(QHBoxLayout())
    self.properties_information_stats.layout().setContentsMargins(0, 0, 0, 0)
    self.properties_information_stats.layout().setSpacing(5)

    self.properties_information_word_counter = QLabel()
    self.properties_information_word_counter.setAlignment(Qt.AlignCenter)
    self.properties_information_word_counter.setObjectName('properties_information_word_counter')
    self.properties_information_stats.layout().addWidget(self.properties_information_word_counter)

    self.properties_information_stats.layout().addSpacing(-5)

    self.properties_information_wpm = QLabel()
    self.properties_information_wpm.setAlignment(Qt.AlignCenter)
    self.properties_information_wpm.setObjectName('properties_information_wpm')
    self.properties_information_stats.layout().addWidget(self.properties_information_wpm)

    self.properties_information_character_counter = QLabel()
    self.properties_information_character_counter.setAlignment(Qt.AlignCenter)
    self.properties_information_character_counter.setObjectName('properties_information_character_counter')
    self.properties_information_stats.layout().addWidget(self.properties_information_character_counter)

    self.properties_information_stats.layout().addSpacing(-5)

    self.properties_information_cps = QLabel()
    self.properties_information_cps.setAlignment(Qt.AlignCenter)
    self.properties_information_cps.setObjectName('properties_information_cps')
    self.properties_information_stats.layout().addWidget(self.properties_information_cps)

    self.properties_information_sub_duration = QLabel()
    self.properties_information_sub_duration.setAlignment(Qt.AlignCenter)
    self.properties_information_sub_duration.setObjectName('properties_information_sub_duration')
    self.properties_information_stats.layout().addWidget(self.properties_information_sub_duration)

    self.properties_information_number_of_lines = QLabel()
    self.properties_information_number_of_lines.setAlignment(Qt.AlignCenter)
    self.properties_information_number_of_lines.setObjectName('properties_information_number_of_lines')
    self.properties_information_stats.layout().addWidget(self.properties_information_number_of_lines)

    self.properties_information_stats.layout().addSpacing(-5)

    self.properties_information_cpl = QLabel()
    self.properties_information_cpl.setObjectName('properties_information_cpl')
    self.properties_information_cpl.setFixedWidth(30)
    self.properties_information_stats.layout().addWidget(self.properties_information_cpl)

    self.properties_information.layout().addWidget(self.properties_information_stats)

    self.properties_information_reason = QLabel()
    self.properties_information_reason.setObjectName('properties_information_reason')
    self.properties_information.layout().addWidget(self.properties_information_reason)

    # self.properties_information.layout().addStretch()

    self.subtitles_panel_simplelist_widget_vbox.addWidget(self.properties_information)

    self.subtitles_panel_simplelist_properties = QFrame()
    self.subtitles_panel_simplelist_properties.setLayout(QVBoxLayout())
    self.subtitles_panel_simplelist_properties.layout().setContentsMargins(20, 0, 0, 0)
    self.subtitles_panel_simplelist_properties.layout().setSpacing(0)

    self.subtitles_panel_simplelist_properties_textedit_line = QHBoxLayout()
    self.subtitles_panel_simplelist_properties_textedit_line.setContentsMargins(0, 0, 0, 0)
    self.subtitles_panel_simplelist_properties_textedit_line.setSpacing(0)

    self.properties_textedit = QTextEdit()
    self.properties_textedit.setObjectName('properties_textedit')
    self.properties_textedit.setFixedHeight(120)
    self.properties_textedit.textChanged.connect(lambda: properties_textedit_changed(self))
    self.subtitles_panel_simplelist_properties_textedit_line.addWidget(self.properties_textedit)

    self.subtitles_panel_simplelist_properties_textedit_timings_column = QVBoxLayout()
    self.subtitles_panel_simplelist_properties_textedit_timings_column.setContentsMargins(0, 0, 0, 0)
    self.subtitles_panel_simplelist_properties_textedit_timings_column.setSpacing(0)

    self.subtitles_panel_simplelist_properties_start_timing_frame = QFrame()
    self.subtitles_panel_simplelist_properties_start_timing_frame.setObjectName('subtitles_panel_simplelist_properties_start_timing_frame')
    self.subtitles_panel_simplelist_properties_start_timing_frame.setProperty('class', 'qframe_timings')
    self.subtitles_panel_simplelist_properties_start_timing_frame.setLayout(QVBoxLayout())
    self.subtitles_panel_simplelist_properties_start_timing_frame.layout().setContentsMargins(2, 2, 2, 2)
    self.subtitles_panel_simplelist_properties_start_timing_frame.layout().setSpacing(0)

    self.subtitles_panel_simplelist_properties_start_timing_label = QLabel('Start')
    self.subtitles_panel_simplelist_properties_start_timing_label.setObjectName('subtitles_panel_simplelist_properties_start_timing_label')
    self.subtitles_panel_simplelist_properties_start_timing_label.setProperty('class', 'properties_timing_labels')
    self.subtitles_panel_simplelist_properties_start_timing_label.setSizePolicy(QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum))
    self.subtitles_panel_simplelist_properties_start_timing_frame.layout().addWidget(self.subtitles_panel_simplelist_properties_start_timing_label)

    self.subtitles_panel_simplelist_properties_start_timing_qlineedit = QLineEdit()
    self.subtitles_panel_simplelist_properties_start_timing_qlineedit.setProperty('class', 'qlineedit_timings')
    self.subtitles_panel_simplelist_properties_start_timing_qlineedit.setObjectName('subtitles_panel_simplelist_properties_start_timing_qlineedit')
    self.subtitles_panel_simplelist_properties_start_timing_qlineedit.setSizePolicy(QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Minimum))
    self.subtitles_panel_simplelist_properties_start_timing_qlineedit.textEdited.connect(lambda: subtitles_panel_simplelist_properties_start_timing_qlineedit_text_edited(self))
    self.subtitles_panel_simplelist_properties_start_timing_frame.layout().addWidget(self.subtitles_panel_simplelist_properties_start_timing_qlineedit)

    self.subtitles_panel_simplelist_properties_textedit_timings_column.addWidget(self.subtitles_panel_simplelist_properties_start_timing_frame)

    self.subtitles_panel_simplelist_properties_textedit_timings_column.addSpacing(-10)
    self.subtitles_panel_simplelist_properties_duration_timing_lock_button = QPushButton()
    self.subtitles_panel_simplelist_properties_duration_timing_lock_button.setObjectName('subtitles_panel_simplelist_properties_duration_timing_lock_button')
    self.subtitles_panel_simplelist_properties_duration_timing_lock_button.setProperty('class', 'subbutton_transparent')
    self.subtitles_panel_simplelist_properties_duration_timing_lock_button.setProperty('borderless_right', 'true')
    self.subtitles_panel_simplelist_properties_duration_timing_lock_button.setCheckable(True)
    self.subtitles_panel_simplelist_properties_duration_timing_lock_button.setFixedSize(QSize(20, 20))
    self.subtitles_panel_simplelist_properties_textedit_timings_column.addWidget(self.subtitles_panel_simplelist_properties_duration_timing_lock_button, 0, Qt.AlignRight)
    self.subtitles_panel_simplelist_properties_textedit_timings_column.addSpacing(-10)

    self.subtitles_panel_simplelist_properties_duration_timing_frame = QFrame()
    self.subtitles_panel_simplelist_properties_duration_timing_frame.setObjectName('subtitles_panel_simplelist_properties_duration_timing_frame')
    self.subtitles_panel_simplelist_properties_duration_timing_frame.setProperty('class', 'qframe_timings')
    self.subtitles_panel_simplelist_properties_duration_timing_frame.setLayout(QVBoxLayout())
    self.subtitles_panel_simplelist_properties_duration_timing_frame.layout().setContentsMargins(2, 2, 2, 2)
    self.subtitles_panel_simplelist_properties_duration_timing_frame.layout().setSpacing(0)

    self.subtitles_panel_simplelist_properties_duration_timing_label = QLabel('Duration')
    self.subtitles_panel_simplelist_properties_duration_timing_label.setObjectName('subtitles_panel_simplelist_properties_duration_timing_label')
    self.subtitles_panel_simplelist_properties_duration_timing_label.setProperty('class', 'properties_timing_labels')
    self.subtitles_panel_simplelist_properties_duration_timing_label.setSizePolicy(QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum))
    self.subtitles_panel_simplelist_properties_duration_timing_frame.layout().addWidget(self.subtitles_panel_simplelist_properties_duration_timing_label)

    self.subtitles_panel_simplelist_properties_duration_timing_qlineedit = QLineEdit()
    self.subtitles_panel_simplelist_properties_duration_timing_qlineedit.setProperty('class', 'qlineedit_timings')
    self.subtitles_panel_simplelist_properties_duration_timing_qlineedit.setObjectName('subtitles_panel_simplelist_properties_duration_timing_qlineedit')
    self.subtitles_panel_simplelist_properties_duration_timing_qlineedit.setSizePolicy(QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Minimum))
    self.subtitles_panel_simplelist_properties_duration_timing_qlineedit.textEdited.connect(lambda: subtitles_panel_simplelist_properties_duration_timing_qlineedit_text_edited(self))
    self.subtitles_panel_simplelist_properties_duration_timing_frame.layout().addWidget(self.subtitles_panel_simplelist_properties_duration_timing_qlineedit)

    self.subtitles_panel_simplelist_properties_textedit_timings_column.addWidget(self.subtitles_panel_simplelist_properties_duration_timing_frame)

    self.subtitles_panel_simplelist_properties_ending_timing_frame = QFrame()
    self.subtitles_panel_simplelist_properties_ending_timing_frame.setObjectName('subtitles_panel_simplelist_properties_ending_timing_frame')
    self.subtitles_panel_simplelist_properties_ending_timing_frame.setProperty('class', 'qframe_timings')
    self.subtitles_panel_simplelist_properties_ending_timing_frame.setLayout(QVBoxLayout())
    self.subtitles_panel_simplelist_properties_ending_timing_frame.layout().setContentsMargins(2, 2, 2, 2)
    self.subtitles_panel_simplelist_properties_ending_timing_frame.layout().setSpacing(0)

    self.subtitles_panel_simplelist_properties_ending_timing_label = QLabel('End')
    self.subtitles_panel_simplelist_properties_ending_timing_label.setObjectName('subtitles_panel_simplelist_properties_ending_timing_label')
    self.subtitles_panel_simplelist_properties_ending_timing_label.setProperty('class', 'properties_timing_labels')
    self.subtitles_panel_simplelist_properties_ending_timing_label.setSizePolicy(QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum))
    self.subtitles_panel_simplelist_properties_ending_timing_frame.layout().addWidget(self.subtitles_panel_simplelist_properties_ending_timing_label)

    self.subtitles_panel_simplelist_properties_ending_timing_qlineedit = QLineEdit()
    self.subtitles_panel_simplelist_properties_ending_timing_qlineedit.setProperty('class', 'qlineedit_timings')
    self.subtitles_panel_simplelist_properties_ending_timing_qlineedit.setObjectName('subtitles_panel_simplelist_properties_ending_timing_qlineedit')
    self.subtitles_panel_simplelist_properties_ending_timing_qlineedit.setSizePolicy(QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Minimum))
    self.subtitles_panel_simplelist_properties_ending_timing_qlineedit.textEdited.connect(lambda: subtitles_panel_simplelist_properties_ending_timing_qlineedit_text_edited(self))
    self.subtitles_panel_simplelist_properties_ending_timing_frame.layout().addWidget(self.subtitles_panel_simplelist_properties_ending_timing_qlineedit)

    self.subtitles_panel_simplelist_properties_textedit_timings_column.addWidget(self.subtitles_panel_simplelist_properties_ending_timing_frame)

    self.subtitles_panel_simplelist_properties_textedit_line.addLayout(self.subtitles_panel_simplelist_properties_textedit_timings_column)

    self.subtitles_panel_simplelist_properties.layout().addLayout(self.subtitles_panel_simplelist_properties_textedit_line)

    self.subtitles_panel_simplelist_properties_buttons_line = QHBoxLayout()
    self.subtitles_panel_simplelist_properties_buttons_line.setContentsMargins(0, 0, 0, 0)
    self.subtitles_panel_simplelist_properties_buttons_line.setSpacing(0)

    self.send_text_to_last_subtitle_button = QPushButton('Send to last')
    self.send_text_to_last_subtitle_button.setObjectName('send_text_to_last_subtitle_button')
    self.send_text_to_last_subtitle_button.setLayout(QHBoxLayout(self.send_text_to_last_subtitle_button))
    self.send_text_to_last_subtitle_button.layout().setContentsMargins(3, 0, 3, 3)
    self.send_text_to_last_subtitle_button.setProperty('class', 'subbutton2_dark')
    self.send_text_to_last_subtitle_button.clicked.connect(lambda: send_text_to_last_subtitle_button_clicked(self))
    self.subtitles_panel_simplelist_properties_buttons_line.addWidget(self.send_text_to_last_subtitle_button)

    self.send_text_to_last_subtitle_and_slice_button = QPushButton()
    self.send_text_to_last_subtitle_and_slice_button.setObjectName('send_text_to_last_subtitle_and_slice_button')
    self.send_text_to_last_subtitle_and_slice_button.setProperty('class', 'subbutton2_dark')
    self.send_text_to_last_subtitle_and_slice_button.clicked.connect(lambda: send_text_to_last_subtitle_and_slice_button_clicked(self))
    self.send_text_to_last_subtitle_and_slice_button.setSizePolicy(QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Minimum))
    self.send_text_to_last_subtitle_button.layout().addWidget(self.send_text_to_last_subtitle_and_slice_button, 0, Qt.AlignLeft)

    self.send_text_to_next_subtitle_button = QPushButton('Send to next')
    self.send_text_to_next_subtitle_button.setObjectName('send_text_to_next_subtitle_button')
    self.send_text_to_next_subtitle_button.setLayout(QHBoxLayout(self.send_text_to_next_subtitle_button))
    self.send_text_to_next_subtitle_button.layout().setContentsMargins(3, 0, 3, 3)
    self.send_text_to_next_subtitle_button.setProperty('class', 'subbutton2_dark')
    self.send_text_to_next_subtitle_button.clicked.connect(lambda: send_text_to_next_subtitle_button_clicked(self))
    self.subtitles_panel_simplelist_properties_buttons_line.addWidget(self.send_text_to_next_subtitle_button)

    self.send_text_to_next_subtitle_and_slice_button = QPushButton()
    self.send_text_to_next_subtitle_and_slice_button.setObjectName('send_text_to_next_subtitle_and_slice_button')
    self.send_text_to_next_subtitle_and_slice_button.setProperty('class', 'subbutton2_dark')
    self.send_text_to_next_subtitle_and_slice_button.clicked.connect(lambda: send_text_to_next_subtitle_and_slice_button_clicked(self))
    self.send_text_to_next_subtitle_and_slice_button.setSizePolicy(QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Minimum))
    self.send_text_to_next_subtitle_button.layout().addWidget(self.send_text_to_next_subtitle_and_slice_button, 0, Qt.AlignRight)

    self.subtitles_panel_simplelist_properties.layout().addLayout(self.subtitles_panel_simplelist_properties_buttons_line)

    self.subtitles_panel_simplelist_widget_vbox.addWidget(self.subtitles_panel_simplelist_properties, 0)

    self.subtitles_panel_stackedwidgets.addWidget(self.subtitles_panel_simplelist_widget)


def add_button(self):
    self.subtitles_panel_widget_button_list = QPushButton()
    self.subtitles_panel_widget_button_list.setObjectName('subtitles_panel_widget_button_list')
    self.subtitles_panel_widget_button_list.setProperty('class', 'subtitles_panel_left_button')
    self.subtitles_panel_widget_button_list.setCheckable(True)
    self.subtitles_panel_widget_button_list.setChecked(True)
    self.subtitles_panel_widget_button_list.setFixedWidth(23)
    self.subtitles_panel_widget_button_list.clicked.connect(lambda vision: subtitles_panel.update_subtitles_panel_widget_vision(self, 'list'))
    self.subtitles_panel_widget_buttons_vbox.addWidget(self.subtitles_panel_widget_button_list)


def subtitles_panel_markdown_qtextedit_cursorpositionchanged(self):
    position = self.subtitles_panel_markdown_qtextedit.textCursor().position()

    cursor = 0
    # markdown_text = ''
    for subtitle in sorted(self.subtitles_list):
        cursor += len(str("{:.3f}".format(subtitle[0])))
        next_index = self.subtitles_list.index(subtitle) + 1
        if not next_index >= len(self.subtitles_list) and not self.subtitles_list[next_index][0] - 0.001 == subtitle[0] + subtitle[1]:
            cursor += len(' - ' + str("{:.3f}".format(subtitle[0] + subtitle[1])))
        cursor += len('\n')

        cursor += len(str(subtitle[2]) + '\n\n')
        if cursor > position:
            self.selected_subtitle = subtitle
            break

    if self.selected_subtitle:
        if not (self.player_widget.position > self.selected_subtitle[0] and self.player_widget.position < self.selected_subtitle[0] + self.selected_subtitle[1]):
            self.player_widget.seek(self.selected_subtitle[0] + (self.selected_subtitle[1] * .5))

    self.timeline.update(self)
    self.timeline.update_scrollbar(self, position='middle')

    # print(self.selected_subtitle)


def subtitles_panel_markdown_qtextedit_textchanged(self):
    subtitles_panel_markdown_qtextedit_update_subtitles_list(self)


def subtitles_panel_markdown_qtextedit_update_subtitles_list(self):
    sub_list = []

    last_time = []
    last_text = ''
    s = 0
    for line in self.subtitles_panel_markdown_qtextedit.toPlainText().split('\n'):
        if all(utils.is_float(t) for t in line.strip().replace(' - ', ' ').split(' ')):
            this_time = [
                float(line.strip().split(' - ')[0])
            ]

            if ' - ' in line.strip():
                this_time.append(float(line.strip().split(' - ')[-1]))
            else:
                this_time.append(0)

            if last_time:
                sub_list.append(
                    [
                        float(last_time[0]),
                        (float(last_time[1]) - float(last_time[0])) if last_time[1] else (float(this_time[0]) - .001 - float(last_time[0])),
                        last_text.strip()
                    ]
                )
                last_text = ''
                s += 1

            last_time = this_time

        else:
            last_text += line + '\n'

    if last_time:
        sub_list.append(
            [
                float(last_time[0]),
                (float(last_time[1]) - float(last_time[0])) if last_time[1] else (float(this_time[0]) - .001 - float(last_time[0])),
                last_text.strip()
            ]
        )
        last_text = ''

    self.selected_subtitle = False

    # Sanitize subtitles so there is no overlaping subtitles?

    self.subtitles_list = sorted(sub_list)

    self.timeline.update(self)


def update_subtitles_panel_qlistwidget(self):
    """Function to update subtitles list widgets"""

    current_sub, index = subtitles.subtitle_under_current_position(subtitles=self.subtitles_list, position=self.player_widget.position)
    if current_sub and not (self.subtitles_panel_qlistwidget.verticalScrollBar().value() + self.subtitles_panel_qlistwidget.verticalScrollBar().pageStep() > index > self.subtitles_panel_qlistwidget.verticalScrollBar().value()):
        self.subtitles_panel_qlistwidget.verticalScrollBar().setValue(index - 1)

    if self.subtitles_list:
        self.subtitles_panel_qlistwidget_model.subtitles = sorted(self.subtitles_list)
        self.subtitles_panel_qlistwidget_model.layoutChanged.emit()

        # counter = 1
        # for sub in sorted(self.subtitles_list):
        #     print(sub)
        #     item, item_widget = generate_qlistwidget_item_widget(self, counter, sub)

        #     self.subtitles_panel_qlistwidget.addItem(item)
        #     self.subtitles_panel_qlistwidget.setItemWidget(item, item_widget)

        #     counter += 1
    if self.selected_subtitle:
        self.subtitles_panel_qlistwidget.setCurrentIndex(self.subtitles_panel_qlistwidget_model.get_index(self.selected_subtitle))

    update_properties_widget(self)


def subtitles_panel_qlistwidget_item_clicked(self):
    """Function to call when a subtitle item on the list is clicked"""
    if self.subtitles_panel_qlistwidget.currentIndex():
        sub_index = self.subtitles_panel_qlistwidget.currentIndex().row()
        self.selected_subtitle = self.subtitles_list[sub_index]

    if self.selected_subtitle:
        self.properties_textedit.blockSignals(True)
        update_properties_widget(self)
        self.properties_textedit.blockSignals(False)

        if not self.selected_subtitle[0] < self.player_widget.position < self.selected_subtitle[0] + self.selected_subtitle[1]:
            self.player_widget.seek(self.selected_subtitle[0] + (self.selected_subtitle[1] * .5))

        self.timeline.update_scrollbar(self, position='middle')


def send_text_to_next_subtitle_button_clicked(self):
    """Function to call when send text to next subtitle is clicked"""
    pos = self.properties_textedit.textCursor().position()
    last_text = self.properties_textedit.toPlainText()[:pos].strip()
    next_text = self.properties_textedit.toPlainText()[pos:].strip()
    subtitles.send_text_to_next_subtitle(subtitles=self.subtitles_list, selected_subtitle=self.selected_subtitle, last_text=last_text, next_text=next_text)
    subtitles_panel.update_subtitles_panel_widget_vision_content(self)
    self.timeline.update(self)

    self.timeline_widget.setFocus(Qt.TabFocusReason)


def send_text_to_last_subtitle_and_slice_button_clicked(self):
    """Function to send text to the last subtitle and slice at the same time"""
    position = self.player_widget.position
    if not self.selected_subtitle[0] + self.selected_subtitle[1] > self.player_widget.position > self.selected_subtitle[0]:
        position = self.selected_subtitle[0] + (self.selected_subtitle[1] * .5)
    subtitles.subtitle_start_to_current_position(subtitles=self.subtitles_list, position=position)
    subtitles.last_end_to_current_position(subtitles=self.subtitles_list, position=position - .001)
    send_text_to_last_subtitle_button_clicked(self)


def send_text_to_next_subtitle_and_slice_button_clicked(self):
    """Function to send text to the next subtitle and slice at the same time"""
    position = self.player_widget.position
    if not self.selected_subtitle[0] + self.selected_subtitle[1] > self.player_widget.position > self.selected_subtitle[0]:
        position = self.selected_subtitle[0] + (self.selected_subtitle[1] * .5)
    subtitles.subtitle_end_to_current_position(subtitles=self.subtitles_list, position=position)
    subtitles.next_start_to_current_position(subtitles=self.subtitles_list, position=position + .001)
    send_text_to_next_subtitle_button_clicked(self)


def send_text_to_last_subtitle_button_clicked(self):
    """Function to call when send text to last subtitle is clicked"""
    pos = self.properties_textedit.textCursor().position()
    last_text = self.properties_textedit.toPlainText()[:pos].strip()
    next_text = self.properties_textedit.toPlainText()[pos:].strip()
    subtitles.send_text_to_last_subtitle(subtitles=self.subtitles_list, selected_subtitle=self.selected_subtitle, last_text=last_text, next_text=next_text)
    subtitles_panel.update_subtitles_panel_widget_vision_content(self)
    self.timeline.update(self)
    self.timeline_widget.setFocus(Qt.TabFocusReason)


def properties_textedit_changed(self):
    """Function to call when properties textedit is changed"""
    old_selected_subtitle = self.selected_subtitle
    if old_selected_subtitle and old_selected_subtitle[2] != self.properties_textedit.toPlainText():
        counter = self.subtitles_list.index(old_selected_subtitle)
        subtitles.change_subtitle_text(subtitles=self.subtitles_list, selected_subtitle=self.subtitles_list[counter], text=self.properties_textedit.toPlainText())
        self.unsaved = True
        subtitles_panel.update_topbar_status(self)
        self.timeline.update(self)
        self.player.update_subtitle_layer(self)
        update_properties_information(self)


def update_properties_information(self):
    if self.selected_subtitle:
        reasons = []
        issues = []

        if self.settings['quality_check'].get('enabled', False):
            _, reasons, issues = quality_check.check_subtitle(self.selected_subtitle, self.settings['quality_check'])

        n_words = len(self.selected_subtitle[2].replace('\n', ' ').split(' '))
        n_char = len(self.selected_subtitle[2].replace('\n', '').replace(' ', ''))

        self.properties_information_word_counter.setStyleSheet('QLabel { background-color: ' + ('#55d43f' if 'wpm' not in issues else '#aa9e1a1a') + '}')
        self.properties_information_wpm.setStyleSheet('QLabel { background-color: ' + ('#bb55d43f' if 'wpm' not in issues else '#bb9e1a1a') + '}')
        self.properties_information_character_counter.setStyleSheet('QLabel { background-color: ' + ('#55d43f' if 'cps' not in issues else '#aa9e1a1a') + '}')
        self.properties_information_cps.setStyleSheet('QLabel { background-color: ' + ('#bb55d43f' if 'cps' not in issues else '#bb9e1a1a') + '}')
        self.properties_information_sub_duration.setStyleSheet('QLabel { background-color: ' + ('#55d43f' if 'duration' not in issues else '#bb9e1a1a') + '}')
        self.properties_information_number_of_lines.setStyleSheet('QLabel { background-color: ' + ('#55d43f' if 'number_of_lines' not in issues else '#bb9e1a1a') + '}')
        self.properties_information_cpl.setStyleSheet('QLabel { background-color: ' + ('#bb55d43f' if 'cpl' not in issues else '#bb9e1a1a') + '}')

        self.properties_information_word_counter.setText(str(n_words))
        self.properties_information_wpm.setText(str(int(n_words / (self.selected_subtitle[1] / 60))))
        self.properties_information_character_counter.setText(str(n_char))
        self.properties_information_cps.setText(str(int(n_char / (self.selected_subtitle[1]))))
        self.properties_information_sub_duration.setText(str(round(self.selected_subtitle[1], 3)))
        self.properties_information_number_of_lines.setText(str(int(len(self.selected_subtitle[2].split('\n')))))

        self.properties_information_reason.setVisible(bool(reasons))
        self.properties_information_reason.setText('\n'.join(reasons))


def update_properties_widget(self):
    """Function to update properties panel widgets"""
    if self.subtitles_panel_stackedwidgets.currentWidget() == self.subtitles_panel_simplelist_widget:
        update_properties_information(self)
        self.subtitles_panel_simplelist_properties.setVisible(bool(self.selected_subtitle))
        self.subtitles_panel_simplelist_properties_duration_timing_lock_button.raise_()

        text = ''
        if self.selected_subtitle:
            text = self.selected_subtitle[2]

            self.subtitles_panel_simplelist_properties_start_timing_qlineedit.setText(utils.get_timeline_time_str(self.selected_subtitle[0], ms=True))
            self.subtitles_panel_simplelist_properties_duration_timing_qlineedit.setText(utils.get_timeline_time_str(self.selected_subtitle[1], ms=True))
            self.subtitles_panel_simplelist_properties_ending_timing_qlineedit.setText(utils.get_timeline_time_str(self.selected_subtitle[0] + self.selected_subtitle[1], ms=True))

        if not self.properties_textedit.hasFocus():
            self.properties_textedit.setText(text)
        self.properties_information_stats.setVisible(bool(self.selected_subtitle) and self.settings.get('quality_check', {}).get('show_statistics', False))


def subtitles_panel_simplelist_properties_start_timing_qlineedit_text_edited(self):
    if self.selected_subtitle:
        if not (self.subtitles_panel_simplelist_properties_start_timing_qlineedit.text() == '' or utils.convert_ffmpeg_timecode_to_seconds(self.subtitles_panel_simplelist_properties_start_timing_qlineedit.text()) > self.selected_subtitle[0] + self.selected_subtitle[1]):
            new_start = utils.convert_ffmpeg_timecode_to_seconds(self.subtitles_panel_simplelist_properties_start_timing_qlineedit.text())
            duration = self.selected_subtitle[1] if self.subtitles_panel_simplelist_properties_duration_timing_lock_button.isChecked() else (self.selected_subtitle[0] + self.selected_subtitle[1] - new_start)
            self.selected_subtitle[0] = new_start
            self.selected_subtitle[1] = duration
    self.timeline.update(self)


def subtitles_panel_simplelist_properties_duration_timing_qlineedit_text_edited(self):
    if self.selected_subtitle:
        if self.subtitles_panel_simplelist_properties_duration_timing_qlineedit.text() == '':
            self.subtitles_panel_simplelist_properties_duration_timing_qlineedit.setText(utils.get_timeline_time_str(self.selected_subtitle[1] - self.selected_subtitle[0], ms=True))
        else:
            self.selected_subtitle[1] = utils.convert_ffmpeg_timecode_to_seconds(self.subtitles_panel_simplelist_properties_duration_timing_qlineedit.text())
    self.timeline.update(self)


def subtitles_panel_simplelist_properties_ending_timing_qlineedit_text_edited(self):
    if self.selected_subtitle:
        if self.subtitles_panel_simplelist_properties_ending_timing_qlineedit.text() == '':
            self.subtitles_panel_simplelist_properties_ending_timing_qlineedit.setText(utils.get_timeline_time_str(self.selected_subtitle[0] + self.selected_subtitle[1], ms=True))
        else:
            self.selected_subtitle[1] = utils.convert_ffmpeg_timecode_to_seconds(self.subtitles_panel_simplelist_properties_ending_timing_qlineedit.text()) - utils.convert_ffmpeg_timecode_to_seconds(self.subtitles_panel_simplelist_properties_starting_timing_qlineedit.text())
    self.timeline.update(self)



"""Subtitles Video panel

"""

from PySide6.QtWidgets import QDoubleSpinBox, QLabel, QPushButton, QSpinBox, QWidget, QVBoxLayout, QCheckBox, QSlider, QGroupBox, QHBoxLayout
from PySide6.QtCore import Qt

from subtitld.interface import global_panel


def load_menu(self):
    """Function to load subtitles panel widgets"""
    self.global_panel_qualitycontrol_menu_button = QPushButton('Quality control')
    self.global_panel_qualitycontrol_menu_button.setCheckable(True)
    self.global_panel_qualitycontrol_menu_button.setProperty('class', 'global_panel_menu')
    self.global_panel_qualitycontrol_menu_button.clicked.connect(lambda: global_panel_menu_changed(self))
    self.global_panel_menu.layout().addWidget(self.global_panel_qualitycontrol_menu_button)


def global_panel_menu_changed(self):
    self.global_panel_qualitycontrol_menu_button.setEnabled(False)
    global_panel.global_panel_menu_changed(self, self.global_panel_qualitycontrol_menu_button, self.global_panel_qualitycontrol_content)


def load_widgets(self):
    """Function to load subtitles panel widgets"""

    self.global_panel_qualitycontrol_content = QWidget()
    self.global_panel_qualitycontrol_content.setLayout(QVBoxLayout())

    self.global_panel_tabwidget_show_statistics_checkbox = QCheckBox('Show statistics')
    self.global_panel_tabwidget_show_statistics_checkbox.setChecked(False)
    self.global_panel_tabwidget_show_statistics_checkbox.stateChanged.connect(lambda: update_quality_settings(self))
    self.global_panel_qualitycontrol_content.layout().addWidget(self.global_panel_tabwidget_show_statistics_checkbox)

    self.global_panel_tabwidget_quality_enable_groupbox = QGroupBox('Quality check')
    self.global_panel_tabwidget_quality_enable_groupbox.setCheckable(True)
    self.global_panel_tabwidget_quality_enable_groupbox.setLayout(QVBoxLayout())
    self.global_panel_tabwidget_quality_enable_groupbox.toggled.connect(lambda: update_quality_settings(self))
    self.global_panel_tabwidget_quality_enable_groupbox.layout().setContentsMargins(10, 10, 10, 10)
    self.global_panel_tabwidget_quality_enable_groupbox.layout().setSpacing(20)

    self.global_panel_tabwidget_quality_readingspeed_vbox = QVBoxLayout()
    self.global_panel_tabwidget_quality_readingspeed_vbox.setContentsMargins(0, 0, 0, 0)
    self.global_panel_tabwidget_quality_readingspeed_vbox.setSpacing(2)

    self.global_panel_tabwidget_quality_readingspeed_label = QLabel('Reading Speed')
    self.global_panel_tabwidget_quality_readingspeed_label.setProperty('class', 'widget_label')
    self.global_panel_tabwidget_quality_readingspeed_vbox.addWidget(self.global_panel_tabwidget_quality_readingspeed_label, 0, Qt.AlignLeft)

    self.global_panel_tabwidget_quality_readingspeed_line = QHBoxLayout()
    self.global_panel_tabwidget_quality_readingspeed_line.setContentsMargins(0, 0, 0, 0)
    self.global_panel_tabwidget_quality_readingspeed_line.setSpacing(5)

    self.global_panel_tabwidget_quality_readingspeed_cps = QSpinBox()
    self.global_panel_tabwidget_quality_readingspeed_cps.setMinimum(0)
    self.global_panel_tabwidget_quality_readingspeed_cps.setMaximum(999)
    self.global_panel_tabwidget_quality_readingspeed_cps.setValue(21)
    self.global_panel_tabwidget_quality_readingspeed_cps.valueChanged.connect(lambda: update_quality_settings(self))
    self.global_panel_tabwidget_quality_readingspeed_line.addWidget(self.global_panel_tabwidget_quality_readingspeed_cps)

    self.global_panel_tabwidget_quality_readingspeed_cps_label = QLabel('Characters per second')
    self.global_panel_tabwidget_quality_readingspeed_cps_label.setProperty('class', 'units_label')
    self.global_panel_tabwidget_quality_readingspeed_line.addWidget(self.global_panel_tabwidget_quality_readingspeed_cps_label, 0, Qt.AlignLeft)

    self.global_panel_tabwidget_quality_readingspeed_wpm = QSpinBox()
    self.global_panel_tabwidget_quality_readingspeed_wpm.setMinimum(0)
    self.global_panel_tabwidget_quality_readingspeed_wpm.setMaximum(999)
    self.global_panel_tabwidget_quality_readingspeed_wpm.setValue(140)
    self.global_panel_tabwidget_quality_readingspeed_wpm.valueChanged.connect(lambda: update_quality_settings(self))
    self.global_panel_tabwidget_quality_readingspeed_line.addWidget(self.global_panel_tabwidget_quality_readingspeed_wpm)

    self.global_panel_tabwidget_quality_readingspeed_wpm_label = QLabel('Words per minute')
    self.global_panel_tabwidget_quality_readingspeed_wpm_label.setProperty('class', 'units_label')
    self.global_panel_tabwidget_quality_readingspeed_line.addWidget(self.global_panel_tabwidget_quality_readingspeed_wpm_label, 0, Qt.AlignLeft)

    self.global_panel_tabwidget_quality_readingspeed_line.addStretch()

    self.global_panel_tabwidget_quality_readingspeed_vbox.addLayout(self.global_panel_tabwidget_quality_readingspeed_line)

    self.global_panel_tabwidget_quality_enable_groupbox.layout().addLayout(self.global_panel_tabwidget_quality_readingspeed_vbox)

    self.global_panel_tabwidget_quality_duration_vbox = QVBoxLayout()
    self.global_panel_tabwidget_quality_duration_vbox.setContentsMargins(0, 0, 0, 0)
    self.global_panel_tabwidget_quality_duration_vbox.setSpacing(2)

    self.global_panel_tabwidget_quality_duration_label = QLabel('Subtitle duration')
    self.global_panel_tabwidget_quality_duration_label.setProperty('class', 'widget_label')
    self.global_panel_tabwidget_quality_duration_vbox.addWidget(self.global_panel_tabwidget_quality_duration_label, 0, Qt.AlignLeft)

    self.global_panel_tabwidget_quality_duration_line = QHBoxLayout()
    self.global_panel_tabwidget_quality_duration_line.setContentsMargins(0, 0, 0, 0)
    self.global_panel_tabwidget_quality_duration_line.setSpacing(5)

    self.global_panel_tabwidget_quality_duration_minimum = QDoubleSpinBox()
    self.global_panel_tabwidget_quality_duration_minimum.setMinimum(0.1)
    self.global_panel_tabwidget_quality_duration_minimum.setMaximum(999.999)
    self.global_panel_tabwidget_quality_duration_minimum.setValue(.7)
    self.global_panel_tabwidget_quality_duration_minimum.valueChanged.connect(lambda: update_quality_settings(self))
    self.global_panel_tabwidget_quality_duration_line.addWidget(self.global_panel_tabwidget_quality_duration_minimum)

    self.global_panel_tabwidget_quality_duration_minimum_label = QLabel('Minimum in seconds')
    self.global_panel_tabwidget_quality_duration_minimum_label.setProperty('class', 'units_label')
    self.global_panel_tabwidget_quality_duration_line.addWidget(self.global_panel_tabwidget_quality_duration_minimum_label)

    self.global_panel_tabwidget_quality_duration_maximum = QDoubleSpinBox()
    self.global_panel_tabwidget_quality_duration_maximum.setMinimum(0.2)
    self.global_panel_tabwidget_quality_duration_maximum.setMaximum(999.999)
    self.global_panel_tabwidget_quality_duration_maximum.setValue(7)
    self.global_panel_tabwidget_quality_duration_maximum.valueChanged.connect(lambda: update_quality_settings(self))
    self.global_panel_tabwidget_quality_duration_line.addWidget(self.global_panel_tabwidget_quality_duration_maximum)

    self.global_panel_tabwidget_quality_duration_maximum_label = QLabel('Maximum in seconds')
    self.global_panel_tabwidget_quality_duration_maximum_label.setProperty('class', 'units_label')
    self.global_panel_tabwidget_quality_duration_line.addWidget(self.global_panel_tabwidget_quality_duration_maximum_label)

    self.global_panel_tabwidget_quality_duration_line.addStretch()

    self.global_panel_tabwidget_quality_duration_vbox.addLayout(self.global_panel_tabwidget_quality_duration_line)

    self.global_panel_tabwidget_quality_enable_groupbox.layout().addLayout(self.global_panel_tabwidget_quality_duration_vbox)

    self.global_panel_tabwidget_quality_lines_vbox = QVBoxLayout()
    self.global_panel_tabwidget_quality_lines_vbox.setContentsMargins(0, 0, 0, 0)
    self.global_panel_tabwidget_quality_lines_vbox.setSpacing(2)

    self.global_panel_tabwidget_quality_lines_label = QLabel('Lines')
    self.global_panel_tabwidget_quality_lines_label.setProperty('class', 'widget_label')
    self.global_panel_tabwidget_quality_lines_vbox.addWidget(self.global_panel_tabwidget_quality_lines_label, 0, Qt.AlignLeft)

    self.global_panel_tabwidget_quality_lines_line = QHBoxLayout()
    self.global_panel_tabwidget_quality_lines_line.setContentsMargins(0, 0, 0, 0)
    self.global_panel_tabwidget_quality_lines_line.setSpacing(5)

    self.global_panel_tabwidget_quality_lines_maximum = QSpinBox()
    self.global_panel_tabwidget_quality_lines_maximum.setMinimum(1)
    self.global_panel_tabwidget_quality_lines_maximum.setMaximum(10)
    self.global_panel_tabwidget_quality_lines_maximum.setValue(2)
    self.global_panel_tabwidget_quality_lines_maximum.valueChanged.connect(lambda: update_quality_settings(self))
    self.global_panel_tabwidget_quality_lines_line.addWidget(self.global_panel_tabwidget_quality_lines_maximum)

    self.global_panel_tabwidget_quality_lines_maximum_label = QLabel('Maximum')
    self.global_panel_tabwidget_quality_lines_maximum_label.setProperty('class', 'units_label')
    self.global_panel_tabwidget_quality_lines_line.addWidget(self.global_panel_tabwidget_quality_lines_maximum_label)

    self.global_panel_tabwidget_quality_lines_maximumcharacters = QSpinBox()
    self.global_panel_tabwidget_quality_lines_maximumcharacters.setMinimum(1)
    self.global_panel_tabwidget_quality_lines_maximumcharacters.setMaximum(999)
    self.global_panel_tabwidget_quality_lines_maximumcharacters.setValue(42)
    self.global_panel_tabwidget_quality_lines_maximumcharacters.valueChanged.connect(lambda: update_quality_settings(self))
    self.global_panel_tabwidget_quality_lines_line.addWidget(self.global_panel_tabwidget_quality_lines_maximumcharacters)

    self.global_panel_tabwidget_quality_lines_maximumcharacters_label = QLabel('Maximum characters per line')
    self.global_panel_tabwidget_quality_lines_maximumcharacters_label.setProperty('class', 'units_label')
    self.global_panel_tabwidget_quality_lines_line.addWidget(self.global_panel_tabwidget_quality_lines_maximumcharacters_label)

    self.global_panel_tabwidget_quality_lines_line.addStretch()

    self.global_panel_tabwidget_quality_lines_vbox.addLayout(self.global_panel_tabwidget_quality_lines_line)

    self.global_panel_tabwidget_quality_lines_vbox.addSpacing(5)

    self.global_panel_tabwidget_quality_prefer_compact_checkbox = QCheckBox('Prefer compact subtitles')
    self.global_panel_tabwidget_quality_prefer_compact_checkbox.stateChanged.connect(lambda: update_quality_settings(self))
    self.global_panel_tabwidget_quality_lines_vbox.addWidget(self.global_panel_tabwidget_quality_prefer_compact_checkbox)

    self.global_panel_tabwidget_quality_lines_vbox.addSpacing(5)

    self.global_panel_tabwidget_quality_balanceratio_checkbox = QCheckBox('Balance line length')
    self.global_panel_tabwidget_quality_balanceratio_checkbox.stateChanged.connect(lambda: update_quality_settings(self))
    self.global_panel_tabwidget_quality_lines_vbox.addWidget(self.global_panel_tabwidget_quality_balanceratio_checkbox)

    self.global_panel_tabwidget_quality_lines_vbox.addSpacing(5)

    self.global_panel_tabwidget_quality_balanceratio_hbox = QHBoxLayout()
    self.global_panel_tabwidget_quality_balanceratio_hbox.setContentsMargins(0, 0, 0, 0)
    self.global_panel_tabwidget_quality_balanceratio_hbox.setSpacing(5)

    self.global_panel_tabwidget_quality_balanceratio_slider = QSlider(orientation=Qt.Horizontal)
    self.global_panel_tabwidget_quality_balanceratio_slider.setObjectName('global_panel_tabwidget_quality_balanceratio_slider')
    self.global_panel_tabwidget_quality_balanceratio_slider.setMinimum(0)
    self.global_panel_tabwidget_quality_balanceratio_slider.setMaximum(100)
    self.global_panel_tabwidget_quality_balanceratio_slider.setValue(50)
    self.global_panel_tabwidget_quality_balanceratio_slider.setMaximumWidth(200)
    self.global_panel_tabwidget_quality_balanceratio_slider.valueChanged.connect(lambda: update_quality_settings(self))
    self.global_panel_tabwidget_quality_balanceratio_hbox.addWidget(self.global_panel_tabwidget_quality_balanceratio_slider)

    self.global_panel_tabwidget_quality_balanceratio_slider_label = QLabel()
    self.global_panel_tabwidget_quality_balanceratio_slider_label.setProperty('class', 'units_label')
    self.global_panel_tabwidget_quality_balanceratio_hbox.addWidget(self.global_panel_tabwidget_quality_balanceratio_slider_label)

    self.global_panel_tabwidget_quality_balanceratio_hbox.addStretch()

    self.global_panel_tabwidget_quality_lines_vbox.addLayout(self.global_panel_tabwidget_quality_balanceratio_hbox)

    self.global_panel_tabwidget_quality_enable_groupbox.layout().addLayout(self.global_panel_tabwidget_quality_lines_vbox)

    self.global_panel_qualitycontrol_content.layout().addWidget(self.global_panel_tabwidget_quality_enable_groupbox)

    self.global_panel_qualitycontrol_content.layout().addStretch()

    self.global_panel_content_stacked_widgets.addWidget(self.global_panel_qualitycontrol_content)

    update_global_panel_qualitycontrol_content_widgets(self)


def update_global_panel_qualitycontrol_content_widgets(self):
    self.global_panel_tabwidget_show_statistics_checkbox.setChecked(self.settings['quality_check'].get('show_statistics', False))
    self.global_panel_tabwidget_quality_enable_groupbox.setChecked(self.settings['quality_check'].get('enabled', False))
    self.global_panel_tabwidget_quality_readingspeed_cps.setValue(self.settings['quality_check'].get('reading_speed_cps', 21))
    self.global_panel_tabwidget_quality_readingspeed_wpm.setValue(self.settings['quality_check'].get('reading_speed_wpm', 140))
    self.global_panel_tabwidget_quality_duration_minimum.setValue(self.settings['quality_check'].get('minimum_duration', .7))
    self.global_panel_tabwidget_quality_duration_maximum.setValue(self.settings['quality_check'].get('maximum_duration', 7))
    self.global_panel_tabwidget_quality_lines_maximum.setValue(self.settings['quality_check'].get('maximum_lines', 2))
    self.global_panel_tabwidget_quality_lines_maximumcharacters.setValue(self.settings['quality_check'].get('maximum_characters_per_line', 42))
    self.global_panel_tabwidget_quality_prefer_compact_checkbox.setChecked(self.settings['quality_check'].get('prefer_compact', False))
    self.global_panel_tabwidget_quality_balanceratio_checkbox.setChecked(self.settings['quality_check'].get('balance_ratio_enabled', False))
    self.global_panel_tabwidget_quality_balanceratio_slider.setValue(self.settings['quality_check'].get('balance_ratio', 50))
    self.global_panel_tabwidget_quality_balanceratio_slider_label.setText('Ratio ({p}% the shortest should be of the largest)'.format(p=self.settings['quality_check'].get('balance_ratio', 50)))


def update_quality_settings(self):
    self.settings['quality_check']['show_statistics'] = self.global_panel_tabwidget_show_statistics_checkbox.isChecked()
    self.settings['quality_check']['enabled'] = self.global_panel_tabwidget_quality_enable_groupbox.isChecked()
    self.settings['quality_check']['reading_speed_cps'] = self.global_panel_tabwidget_quality_readingspeed_cps.value()
    self.settings['quality_check']['reading_speed_wpm'] = self.global_panel_tabwidget_quality_readingspeed_wpm.value()
    self.settings['quality_check']['minimum_duration'] = self.global_panel_tabwidget_quality_duration_minimum.value()
    self.settings['quality_check']['maximum_duration'] = self.global_panel_tabwidget_quality_duration_maximum.value()
    self.settings['quality_check']['maximum_lines'] = self.global_panel_tabwidget_quality_lines_maximum.value()
    self.settings['quality_check']['maximum_characters_per_line'] = self.global_panel_tabwidget_quality_lines_maximumcharacters.value()
    self.settings['quality_check']['prefer_compact'] = self.global_panel_tabwidget_quality_prefer_compact_checkbox.isChecked()
    self.settings['quality_check']['balance_ratio_enabled'] = self.global_panel_tabwidget_quality_balanceratio_checkbox.isChecked()
    self.settings['quality_check']['balance_ratio'] = self.global_panel_tabwidget_quality_balanceratio_slider.value()
    update_global_panel_qualitycontrol_content_widgets(self)



"""Subtitles Video panel

"""
import os
import subprocess

from PySide6.QtWidgets import QLabel, QComboBox, QPushButton, QFileDialog, QSpinBox, QColorDialog, QWidget, QTableWidgetItem, QVBoxLayout, QStackedWidget
from PySide6.QtCore import QMargins, QSize, QThread, Signal, Qt
from PySide6.QtGui import QBrush, QColor, QFont, QFontDatabase, QPainter, QPen
from subtitld.interface import global_panel

from subtitld.modules.paths import LIST_OF_SUPPORTED_EXPORT_EXTENSIONS, STARTUPINFO, FFMPEG_EXECUTABLE, path_tmp
from subtitld.modules.shortcuts import shortcuts_dict
from subtitld.modules import file_io
from subtitld.modules import utils


class ThreadGeneratedBurnedVideo(QThread):
    """Thread to generate burned video"""
    response = Signal(str)
    commands = []

    def run(self):
        """Run function of thread to generate burned video"""
        if self.commands:
            proc = subprocess.Popen(self.commands, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True, startupinfo=STARTUPINFO, bufsize=4096)
            number_of_steps = 0.001
            current_step = 0.0
            while proc.poll() is None:
                # for output in proc.stdout.read().decode().split('\n'):
                output = proc.stdout.readline()
                if 'Duration: ' in output:
                    duration = int(utils.convert_ffmpeg_timecode_to_seconds(output.split('Duration: ', 1)[1].split(',', 1)[0]))
                    if duration > number_of_steps:
                        number_of_steps = duration
                if output[:6] == 'frame=':
                    current_step = int(utils.convert_ffmpeg_timecode_to_seconds(output.split('time=', 1)[1].split(' ', 1)[0]))

                self.response.emit(str(current_step) + '|' + str(number_of_steps))
            self.response.emit('end')


def load_menu(self):
    """Function to load subtitles panel widgets"""
    self.global_panel_export_menu_button = QPushButton('Export')
    self.global_panel_export_menu_button.setCheckable(True)
    self.global_panel_export_menu_button.setProperty('class', 'global_panel_menu')
    self.global_panel_export_menu_button.clicked.connect(lambda: global_panel_menu_changed(self))
    self.global_panel_menu.layout().addWidget(self.global_panel_export_menu_button)


def global_panel_menu_changed(self):
    self.global_panel_export_menu_button.setEnabled(False)
    global_panel.global_panel_menu_changed(self, self.global_panel_export_menu_button, self.global_panel_export_content)


def load_widgets(self):
    """Function to load subtitles panel widgets"""

    self.global_panel_export_content = QWidget()
    self.global_panel_export_content.setLayout(QVBoxLayout())
    self.global_panel_export_content.layout().setContentsMargins(0, 0, 0, 0)

    self.global_panel_export_types_combobox = QComboBox()
    self.global_panel_export_types_combobox.addItems(
        [
            'txt',
            'mp4'
        ]
    )
    self.global_panel_export_types_combobox.activated.connect(lambda: global_panel_export_types_combobox_activated(self))
    self.global_panel_export_content.layout().addWidget(self.global_panel_export_types_combobox, 0, Qt.AlignLeft)

    self.global_panel_export_content_stackedwidgets = QStackedWidget()

    self.global_panel_export_content_txt_panel = QWidget()
    self.global_panel_export_content_txt_panel.setLayout(QVBoxLayout())

    self.global_panel_export_content_txt_panel_export_button = QPushButton(self.tr('Export').upper())
    self.global_panel_export_content_txt_panel_export_button.setProperty('class', 'button')
    self.global_panel_export_content_txt_panel_export_button.clicked.connect(lambda: global_subtitlesvideo_export_button_clicked(self))
    self.global_panel_export_content_txt_panel.layout().addWidget(self.global_panel_export_content_txt_panel_export_button)

    self.global_panel_export_content_stackedwidgets.addWidget(self.global_panel_export_content_txt_panel)

    self.global_panel_export_content_mp4_panel = QWidget()
    self.global_panel_export_content_mp4_panel.setLayout(QVBoxLayout())

    self.global_subtitlesvideo_export_button = QPushButton(self.tr('Export').upper())
    self.global_subtitlesvideo_export_button.setProperty('class', 'button')
    self.global_subtitlesvideo_export_button.clicked.connect(lambda: global_subtitlesvideo_export_button_clicked(self))
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_export_button)

    self.global_subtitlesvideo_video_burn_label = QLabel(self.tr('Export burned video').upper())
    self.global_subtitlesvideo_video_burn_label.setStyleSheet('QLabel { font-size:14px; font-weight:bold; }')
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_label)

    self.global_subtitlesvideo_video_burn_fontname_label = QLabel(self.tr('Font name').upper())
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_fontname_label)

    fonts = QFontDatabase().families()
    self.global_subtitlesvideo_video_burn_fontname = QComboBox(parent=self.global_panel_export_content)
    self.global_subtitlesvideo_video_burn_fontname.setProperty('class', 'button')
    self.global_subtitlesvideo_video_burn_fontname.addItems(fonts)
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_fontname)
    # self.global_subtitlesvideo_video_burn_fontname.activated.connect(lambda: global_subtitlesvideo_save_as_combobox_activated(self))

    self.global_subtitlesvideo_video_burn_fontsize_label = QLabel(self.tr('Font size').upper())
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_fontsize_label)

    self.global_subtitlesvideo_video_burn_fontsize = QSpinBox(parent=self.global_panel_export_content)
    self.global_subtitlesvideo_video_burn_fontsize.setMinimum(8)
    self.global_subtitlesvideo_video_burn_fontsize.setMaximum(200)
    self.global_subtitlesvideo_video_burn_fontsize.setValue(20)
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_fontsize)

    self.global_subtitlesvideo_video_burn_shadowdistance_label = QLabel(self.tr('Shadow distance').upper())
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_shadowdistance_label)

    self.global_subtitlesvideo_video_burn_shadowdistance = QSpinBox(parent=self.global_panel_export_content)
    self.global_subtitlesvideo_video_burn_shadowdistance.setMinimum(0)
    self.global_subtitlesvideo_video_burn_shadowdistance.setMaximum(20)
    self.global_subtitlesvideo_video_burn_shadowdistance.setValue(1)
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_shadowdistance)

    self.global_subtitlesvideo_video_burn_outline_label = QLabel(self.tr('Outline').upper())
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_outline_label)

    self.global_subtitlesvideo_video_burn_outline = QSpinBox(parent=self.global_panel_export_content)
    self.global_subtitlesvideo_video_burn_outline.setMinimum(0)
    self.global_subtitlesvideo_video_burn_outline.setMaximum(20)
    self.global_subtitlesvideo_video_burn_outline.setValue(2)
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_outline)

    self.global_subtitlesvideo_video_burn_marvinv_label = QLabel(self.tr('Margin from bottom').upper())
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_marvinv_label)

    self.global_subtitlesvideo_video_burn_marvinv = QSpinBox(parent=self.global_panel_export_content)
    self.global_subtitlesvideo_video_burn_marvinv.setMinimum(0)
    self.global_subtitlesvideo_video_burn_marvinv.setMaximum(500)
    self.global_subtitlesvideo_video_burn_marvinv.setValue(20)
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_marvinv)

    self.global_subtitlesvideo_video_burn_marvinl_label = QLabel(self.tr('Margin from left').upper())
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_marvinl_label)

    self.global_subtitlesvideo_video_burn_marvinl = QSpinBox(parent=self.global_panel_export_content)
    self.global_subtitlesvideo_video_burn_marvinl.setMinimum(0)
    self.global_subtitlesvideo_video_burn_marvinl.setMaximum(500)
    self.global_subtitlesvideo_video_burn_marvinl.setValue(50)
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_marvinl)

    self.global_subtitlesvideo_video_burn_marvinr_label = QLabel(self.tr('Margin from right').upper())
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_marvinr_label)

    self.global_subtitlesvideo_video_burn_marvinr = QSpinBox(parent=self.global_panel_export_content)
    self.global_subtitlesvideo_video_burn_marvinr.setMinimum(0)
    self.global_subtitlesvideo_video_burn_marvinr.setMaximum(500)
    self.global_subtitlesvideo_video_burn_marvinr.setValue(50)
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_marvinr)

    self.global_subtitlesvideo_video_burn_pcolor_selected_color = '#ffffff'

    self.global_subtitlesvideo_video_burn_pcolor_label = QLabel(self.tr('Font color').upper())
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_pcolor_label)

    self.global_subtitlesvideo_video_burn_pcolor = QPushButton(parent=self.global_panel_export_content)
    self.global_subtitlesvideo_video_burn_pcolor.clicked.connect(lambda: global_subtitlesvideo_video_burn_pcolor_clicked(self))
    self.global_subtitlesvideo_video_burn_pcolor.setStyleSheet('background-color:' + self.global_subtitlesvideo_video_burn_pcolor_selected_color)
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_pcolor)

    self.global_subtitlesvideo_video_burn_convert = QPushButton(self.tr('Generate video').upper())
    self.global_subtitlesvideo_video_burn_convert.setProperty('class', 'button_dark')
    self.global_subtitlesvideo_video_burn_convert.clicked.connect(lambda: global_subtitlesvideo_video_burn_convert_clicked(self))
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_burn_convert)

    self.global_subtitlesvideo_video_generate_transparent_video_button = QPushButton(self.tr('Generate transparent video').upper())
    self.global_subtitlesvideo_video_generate_transparent_video_button.setProperty('class', 'button_dark')
    self.global_subtitlesvideo_video_generate_transparent_video_button.clicked.connect(lambda: global_subtitlesvideo_video_generate_transparent_video_button_clicked(self))
    self.global_panel_export_content_mp4_panel.layout().addWidget(self.global_subtitlesvideo_video_generate_transparent_video_button)
    # self.global_subtitlesvideo_video_generate_transparent_video_button.setVisible(False)

    def thread_generated_burned_video_ended(response):
        if '|' in response:
            self.global_subtitlesvideo_video_burn_convert.setText('GENERATING... (' + str(int((float(response.split('|')[0]) / float(response.split('|')[1])) * 100)) + '%)')
        elif 'end' in response:
            self.global_subtitlesvideo_video_burn_convert.setText(self.tr('Generate video').upper())
            self.global_subtitlesvideo_video_burn_convert.setEnabled(True)

    self.thread_generated_burned_video = ThreadGeneratedBurnedVideo(self)
    self.thread_generated_burned_video.response.connect(thread_generated_burned_video_ended)

    self.global_panel_export_content_stackedwidgets.addWidget(self.global_panel_export_content_mp4_panel)

    self.global_panel_export_content.layout().addWidget(self.global_panel_export_content_stackedwidgets)

    self.global_panel_content_stacked_widgets.addWidget(self.global_panel_export_content)


def global_panel_tabwidget_shortkeys_table_update(self):
    """Function to update subtitlesvideo panel shorkeys table"""
    self.global_panel_tabwidget_shortkeys_table.clear()
    self.global_panel_tabwidget_shortkeys_table.setRowCount(len(shortcuts_dict))
    inverted_shortcuts_dict = {value: key for key, value in shortcuts_dict.items()}
    i = 0
    for item in shortcuts_dict:
        item_name = QTableWidgetItem(shortcuts_dict[item])
        self.global_panel_tabwidget_shortkeys_table.setItem(i, 0, item_name)
        item_name = QTableWidgetItem(self.settings['shortcuts'].get(inverted_shortcuts_dict[shortcuts_dict[item]], [''])[0])
        self.global_panel_tabwidget_shortkeys_table.setItem(i, 1, item_name)
        i += 1


def global_subtitlesvideo_video_burn_convert_clicked(self):
    """Function to generate buned video"""
    suggested_path = os.path.dirname(self.video_metadata['filepath'])
    extformat = os.path.basename(self.video_metadata['filepath']).rsplit('.', 1)[1]
    save_formats = self.tr('Video file') + ' (.' + extformat + ')'
    suggested_name = os.path.basename(self.video_metadata['filepath']).rsplit('.', 1)[0] + '_subtitled.' + extformat

    generated_video_filepath = QFileDialog.getSaveFileName(parent=self, caption=self.tr('Select the subtitle file'), dir=os.path.join(suggested_path, suggested_name), filter=save_formats)[0]

    if generated_video_filepath:
        file_io.save_file(os.path.join(path_tmp, 'subtitle.srt'), self.subtitles_list, subtitle_format='SRT', language='en')

        vf_string = 'subtitles=filename=' + os.path.join(path_tmp, 'subtitle.srt').replace('\\', '\\\\\\\\').replace(':', '\\\:') + ":force_style='"
        vf_string += 'FontName=' + self.global_subtitlesvideo_video_burn_fontname.currentText() + ','
        vf_string += 'FontSize=' + str(self.global_subtitlesvideo_video_burn_fontsize.value()) + ','
        vf_string += 'Shadow=' + str(self.global_subtitlesvideo_video_burn_shadowdistance.value()) + ','
        vf_string += 'Outline=' + str(self.global_subtitlesvideo_video_burn_outline.value()) + ','
        vf_string += 'MarginV=' + str(self.global_subtitlesvideo_video_burn_marvinv.value()) + ','
        vf_string += 'MarginL=' + str(self.global_subtitlesvideo_video_burn_marvinl.value()) + ','
        vf_string += 'MarginR=' + str(self.global_subtitlesvideo_video_burn_marvinr.value()) + ','
        pcolor = self.global_subtitlesvideo_video_burn_pcolor_selected_color.replace('#', '&H')
        pcolor = pcolor[:2] + pcolor[-2:] + pcolor[-4:-2] + pcolor[-6:-4]
        vf_string += 'PrimaryColour=' + pcolor + "'"

        commands = [
            FFMPEG_EXECUTABLE,
            '-i', self.video_metadata['filepath'],
            '-y',
            '-vf',
            vf_string,
            '-crf', '15',
            generated_video_filepath
        ]

        self.thread_generated_burned_video.commands = commands
        self.thread_generated_burned_video.start()
        self.global_subtitlesvideo_video_burn_convert.setEnabled(False)


def global_subtitlesvideo_video_burn_pcolor_clicked(self):
    """Function to change color"""
    color = QColorDialog().getColor(options=QColorDialog.DontUseNativeDialog)
    if color.isValid():
        self.global_subtitlesvideo_video_burn_pcolor_selected_color = color.name()
    self.global_subtitlesvideo_video_burn_pcolor.setStyleSheet('background-color:' + self.global_subtitlesvideo_video_burn_pcolor_selected_color)


def global_subtitlesvideo_video_generate_transparent_video_button_clicked(self):
    suggested_path = os.path.dirname(self.video_metadata['filepath'])
    extformat = 'mov'  # os.path.basename(self.video_metadata['filepath']).rsplit('.', 1)[1]
    save_formats = self.tr('Video file') + ' (.' + extformat + ')'
    suggested_name = os.path.basename(self.video_metadata['filepath']).rsplit('.', 1)[0] + '_subtitled.' + extformat

    generated_video_filepath = QFileDialog.getSaveFileName(parent=self, caption=self.tr('Select the subtitle file'), dir=os.path.join(suggested_path, suggested_name), filter=save_formats)[0]

    # print(path_tmp)

    if generated_video_filepath:
        class layerWidget(QWidget):
            subtitle_text = ''
            font = 'Ubuntu'
            fontsize = 18

            def paintEvent(canvas, paintEvent):
                painter = QPainter(canvas)
                painter.setRenderHint(QPainter.Antialiasing)

                if canvas.subtitle_text:
                    font = QFont(canvas.font)
                    font.setPointSize(canvas.fontsize)
                    painter.setFont(font)

                    text_rect = painter.boundingRect(canvas.width() * .08, canvas.height() * .08, canvas.width() * .84, canvas.height() * .84, Qt.AlignBottom | Qt.AlignHCenter | Qt.TextWordWrap, canvas.subtitle_text)

                    painter.setPen(Qt.NoPen)
                    painter.setBrush(QBrush(QColor('#cc000000')))
                    painter.drawRoundedRect(text_rect.marginsAdded(QMargins(10, 10, 10, 10)), 5, 5)

                    painter.setPen(QPen(QColor('#fff')))
                    painter.drawText(text_rect.adjusted(0, 2, 0, 2), Qt.AlignCenter | Qt.TextWordWrap, canvas.subtitle_text)

                painter.end()

        layer = layerWidget(self)
        layer.setVisible(False)
        layer.setStyleSheet('background: transparent;')
        layer.resize(QSize(self.video_metadata.get('width', 1920), self.video_metadata.get('height', 1920)))
        layer.font = self.global_subtitlesvideo_video_burn_fontname.currentText()
        layer.fontsize = self.global_subtitlesvideo_video_burn_fontsize.value()
        final_text = ''

        i = 0
        last_position = .0
        # if not self.subtitles_list[0][0] == .0:
        #     filename = os.path.join(path_tmp, 'empty.png')
        #     layer.subtitle_text = ''
        #     layer.grab().save(filename, 'PNG')

        #     final_text += "file '" + filename + "'\n"
        #     final_text += 'duration {}\n'.format(self.subtitles_list[0][0])

        #     i += 1

        for subtitle in self.subtitles_list:
            if not last_position + .001 > subtitle[0] - .001:
                filename = os.path.join(path_tmp, 'empty.png')
                layer.subtitle_text = ''
                layer.grab().save(filename, 'PNG')

                final_text += "file '" + filename + "'\n"
                final_text += 'duration {}\n'.format(subtitle[0] - last_position)

            filename = os.path.join(path_tmp, '{}.png'.format(i))
            layer.subtitle_text = subtitle[2]
            layer.grab().save(filename, 'PNG')

            final_text += "file '" + filename + "'\n"
            final_text += 'duration {}\n'.format(subtitle[1])

            last_position = subtitle[0] + subtitle[1]
            i += 1
            print(i)

        if not self.subtitles_list[-1][0] + self.subtitles_list[-1][1] == self.video_metadata.get('duration', 60.0):
            filename = os.path.join(path_tmp, 'empty.png')
            layer.subtitle_text = ''
            layer.grab().save(filename, 'PNG')

            final_text += "file '" + filename + "'\n"
            final_text += 'duration {}\n'.format(self.subtitles_list[0][0])
            final_text += "file '" + filename + "'\n"

        open(os.path.join(path_tmp, 'subtitles.txt'), 'w').write(final_text)
        subprocess.Popen([FFMPEG_EXECUTABLE, '-y', '-f', 'concat', '-safe', '0', '-i', os.path.join(path_tmp, 'subtitles.txt'), '-r', str(self.video_metadata.get('framerate', 24)), '-c:v', 'qtrle', '-an', generated_video_filepath], startupinfo=STARTUPINFO, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT).stdout.read()


def global_subtitlesvideo_export_button_clicked(self):
    """Function to export subtitles"""
    suggested_path = os.path.dirname(self.video_metadata['filepath'])
    suggested_name = os.path.basename(self.video_metadata['filepath']).rsplit('.', 1)[0] + '.txt'

    supported_export_files = ';;'.join(['{description} ({extension})'.format(extension=' '.join(['.{ext}'.format(ext=ext) for ext in LIST_OF_SUPPORTED_EXPORT_EXTENSIONS[export_format]['extensions']]), description=LIST_OF_SUPPORTED_EXPORT_EXTENSIONS[export_format]['description']) for export_format in LIST_OF_SUPPORTED_EXPORT_EXTENSIONS])

    filedialog = QFileDialog.getSaveFileName(parent=self, caption=self.tr('Export to file'), dir=os.path.join(suggested_path, suggested_name), filter=supported_export_files)

    if filedialog[0] and filedialog[1]:
        filename = filedialog[0]
        exts = []
        for extformat in filedialog[1].split('(', 1)[1].split(')', 1)[0].split('*'):
            if extformat:
                exts.append(extformat.strip())
        if not filename.endswith(tuple(exts)):
            filename += exts[0]
        format_to_export = filedialog[1].rsplit('(', 1)[-1].split(')', 1)[0]

        file_io.export_file(filename=filename, subtitles_list=self.subtitles_list, export_format=format_to_export)


def global_panel_export_types_combobox_activated(self):
    if self.global_panel_export_types_combobox.currentText() == 'txt':
        self.global_panel_export_content_stackedwidgets.setCurrentWidget(self.global_panel_export_content_txt_panel)
    elif self.global_panel_export_types_combobox.currentText() == 'mp4':
        self.global_panel_export_content_stackedwidgets.setCurrentWidget(self.global_panel_export_content_mp4_panel)




"""History module (Ctrl+Z)

"""

import copy

ALL_HISTORY = []
REDO_HISTORY = []


def history_append(subtitles):
    """Append subtitles list to history"""
    ALL_HISTORY.append(copy.deepcopy(subtitles))
    REDO_HISTORY.clear()


def history_undo(actual_subtitles):
    """Revert to last subtitle on the history list"""
    if ALL_HISTORY:
        REDO_HISTORY.append(copy.deepcopy(actual_subtitles))
        actual_subtitles.clear()
        actual_subtitles.extend(copy.deepcopy(ALL_HISTORY.pop()))


def history_redo(actual_subtitles):
    """Redo subtitle on the history list"""
    if REDO_HISTORY:
        actual_subtitles.clear()
        actual_subtitles.extend(copy.deepcopy(REDO_HISTORY.pop()))



"""Subtitles Video panel

"""

import subprocess
from google_trans_new import google_translator

from PySide6.QtWidgets import QComboBox, QPushButton, QWidget, QMessageBox, QGridLayout
from PySide6.QtCore import QThread, Signal, Qt

from subtitld.modules.paths import LANGUAGE_DICT_LIST, STARTUPINFO
from subtitld.modules import utils
from subtitld.interface import global_panel


LANGUAGE_DESCRIPTIONS = LANGUAGE_DICT_LIST.keys()


class ThreadGeneratedBurnedVideo(QThread):
    """Thread to generate burned video"""
    response = Signal(str)
    commands = []

    def run(self):
        """Run function of thread to generate burned video"""
        if self.commands:
            proc = subprocess.Popen(self.commands, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True, startupinfo=STARTUPINFO, bufsize=4096)
            number_of_steps = 0.001
            current_step = 0.0
            while proc.poll() is None:
                # for output in proc.stdout.read().decode().split('\n'):
                output = proc.stdout.readline()
                if 'Duration: ' in output:
                    duration = int(utils.convert_ffmpeg_timecode_to_seconds(output.split('Duration: ', 1)[1].split(',', 1)[0]))
                    if duration > number_of_steps:
                        number_of_steps = duration
                if output[:6] == 'frame=':
                    current_step = int(utils.convert_ffmpeg_timecode_to_seconds(output.split('time=', 1)[1].split(' ', 1)[0]))

                self.response.emit(str(current_step) + '|' + str(number_of_steps))
            self.response.emit('end')


def load_menu(self):
    """Function to load subtitles panel widgets"""
    self.global_panel_translation_menu_button = QPushButton('Translation')
    self.global_panel_translation_menu_button.setCheckable(True)
    self.global_panel_translation_menu_button.setProperty('class', 'global_panel_menu')
    self.global_panel_translation_menu_button.clicked.connect(lambda: global_panel_menu_changed(self))
    self.global_panel_menu.layout().addWidget(self.global_panel_translation_menu_button)


def global_panel_menu_changed(self):
    self.global_panel_translation_menu_button.setEnabled(False)
    global_panel.global_panel_menu_changed(self, self.global_panel_translation_menu_button, self.global_panel_translation_content)


def load_widgets(self):
    """Function to load subtitles panel widgets"""

    self.global_panel_translation_content = QWidget()
    self.global_panel_translation_content.setLayout(QGridLayout())

    self.global_subtitlesvideo_autosync_lang_combobox = QComboBox()
    self.global_subtitlesvideo_autosync_lang_combobox.setProperty('class', 'button')
    self.global_subtitlesvideo_autosync_lang_combobox.addItems(LANGUAGE_DESCRIPTIONS)
    self.global_panel_translation_content.layout().addWidget(self.global_subtitlesvideo_autosync_lang_combobox, 0, 0, Qt.AlignTop)

    self.global_subtitlesvideo_translate_button = QPushButton(self.tr('Translate').upper())
    self.global_subtitlesvideo_translate_button.setProperty('class', 'button')
    self.global_subtitlesvideo_translate_button.clicked.connect(lambda: global_subtitlesvideo_translate_button_clicked(self))
    self.global_panel_translation_content.layout().addWidget(self.global_subtitlesvideo_translate_button, 0, 1, Qt.AlignTop)

    self.global_panel_content_stacked_widgets.addWidget(self.global_panel_translation_content)


def global_subtitlesvideo_translate_button_clicked(self):
    """Function to translate subtitles"""
    run_command = False

    if bool(self.subtitles_list):
        are_you_sure_message = QMessageBox(self)
        are_you_sure_message.setWindowTitle(self.tr('Are you sure?'))
        are_you_sure_message.setText(self.tr('This will overwrite your actual subtitle set. New text will be applied. Are you sure you want to replace your actual subtitles?'))
        are_you_sure_message.addButton(self.tr('Yes'), QMessageBox.AcceptRole)
        are_you_sure_message.addButton(self.tr('No'), QMessageBox.RejectRole)
        ret = are_you_sure_message.exec_()

        if ret == QMessageBox.AcceptRole:
            run_command = True
    else:
        run_command = True

    if run_command:

        language = LANGUAGE_DICT_LIST[self.global_subtitlesvideo_autosync_lang_combobox.currentText()].split('-')[0]
        translator = google_translator()  # translator(service_urls=['translate.googleapis.com', 'translate.google.com','translate.google.co.kr'])
        for subtitle in self.subtitles_list:
            subtitle[2] = translator.translate(subtitle[2].replace('\n', ' ').replace('  ', ' '), lang_tgt=language)

        # update_widgets(self)



"""Module to check subtitles quality

"""

def check_subtitle(subtitle=[], settings={}):
    """Function to check individual subtitle"""
    approved = True
    reasons = []
    issues = []
    if subtitle and subtitle[1]:
        if len(subtitle[2].replace('\n', '')) / subtitle[1] > settings.get('reading_speed_cps', 21):
            approved = False
            reasons.append('Reading speed: {ars} characters per second. It should be {rs} per second.'.format(rs=settings.get('reading_speed_cps', 21), ars=round(len(subtitle[2].replace('\n', '')) / subtitle[1], 2)))
            issues.append('cps')
        if len(subtitle[2].replace('\n', '').split(' ')) / (subtitle[1] / 60) > settings.get('reading_speed_wpm', 140):
            approved = False
            reasons.append('Reading speed: {ars} words per minute. It should be {rs} per minute.'.format(rs=settings.get('reading_speed_wpm', 140), ars=round(len(subtitle[2].replace('\n', '').split(' ')) / (subtitle[1] / 60))))
            issues.append('wpm')
        if subtitle[1] < settings.get('minimum_duration', .7):
            approved = False
            reasons.append('The duration of this subtitle, {ad}, is less than the minimum of {d}.'.format(d=settings.get('minimum_duration', .7), ad=round(subtitle[1], 2)))
            issues.append('duration')
        if subtitle[1] > settings.get('maximum_duration', 7):
            approved = False
            reasons.append('The duration of this subtitle, {ad}, is more than the maximum of {d}.'.format(d=settings.get('maximum_duration', 7), ad=round(subtitle[1], 2)))
            issues.append('duration')
        if len(subtitle[2].split('\n')) > settings.get('maximum_lines', 2):
            approved = False
            reasons.append('There are {al} lines, more than the maximum of {l}.'.format(l=settings.get('maximum_lines', 2), al=round(len(subtitle[2].split('\n')), 2)))
            issues.append('number_of_lines')
        for line in subtitle[2].split('\n'):
            if len(line) > settings.get('maximum_characters_per_line', 42):
                approved = False
                reasons.append('There are more than {c} characters per line ({ac}).'.format(c=settings.get('maximum_characters_per_line', 42), ac=len(line)))
                issues.append('cpl')
                break
        if len(subtitle[2].split('\n')) > 1:
            if settings.get('prefer_compact', False):
                if len(subtitle[2].replace('\n', '')) < settings.get('maximum_characters_per_line', 42):
                    approved = False
                    reasons.append('The lines should be joined as there are less than {c} characters in total.'.format(c=settings.get('maximum_characters_per_line', 42)))
            if settings.get('balance_ratio_enabled', False):
                if len(min(subtitle[2].split('\n'), key=len)) < len(max(subtitle[2].split('\n'), key=len))*(settings.get('balance_ratio', 50)/100):
                    approved = False
                    reasons.append('The shortest line length is less than {p}% the largest.'.format(p=settings.get('balance_ratio', 50)))

    return approved, reasons, issues



"""Config functions. Load and save files.

"""

import os
import json

from subtitld.modules.paths import PATH_SUBTITLD_USER_CONFIG_FILE


def load(config_file_path=False):
    """Config load function. Provide a file path with 'config_file_path'.
    It will return a dict with the settings.
    """
    config = {}
    if config_file_path and os.path.isfile(config_file_path):
        with open(config_file_path) as fileobj:
            config = json.load(fileobj)

    if not config.get('recent_files', False):
        config['recent_files'] = {}

    if not config.get('shortcuts', False):
        config['shortcuts'] = {}

    if not config.get('safety_margins', False):
        config['safety_margins'] = {}

    if not config.get('autosave', False):
        config['autosave'] = {}

    if not config.get('timeline', False):
        config['timeline'] = {}

    if not config.get('videoplayer', False):
        config['videoplayer'] = {}

    if not config.get('quality_check', False):
        config['quality_check'] = {}

    if not config.get('default_values', False):
        config['default_values'] = {}

    return config


def save(config=False, config_file_path=False):
    """Config save function. Provide a dict and
    a file path with 'config_file_path'.
    """
    if config:
        if not config_file_path:
            config_file_path = os.path.join(PATH_SUBTITLD_USER_CONFIG_FILE)
        with open(config_file_path, 'w') as fileobj:
            json.dump(config, fileobj, indent=4, sort_keys=True, ensure_ascii=False)



"""Waveform module

"""

import os
import subprocess
import json
from PySide6.QtCore import QPointF, QThread, Signal, Qt
from PySide6.QtGui import QColor, QPainter, QPen, QPixmap, QPolygonF
import numpy
import ffms2
import asyncio

from subtitld.modules.paths import STARTUPINFO, FFMPEG_EXECUTABLE, path_tmp, FFPROBE_EXECUTABLE


def return_audio_numpy(self, audionp):
    """Function to return waveform's numpy"""
    self.video_metadata['waveform'][0] = audionp


def ffmpeg_load_audio(filepath, samplerate=48000, mono=True, normalize=True, in_type=numpy.int16, out_type=numpy.float32):
    """Function to call ffmpeg and return numpy"""
    channels = 1 if mono else 2
    format_strings = {
        numpy.float64: 'f64le',
        numpy.float32: 'f32le',
        numpy.int16: 's16le',
        numpy.int32: 's32le',
        numpy.uint32: 'u32le'
    }
    print('ffmpeg load audio started')
    format_string = format_strings[in_type]
    command = [
        FFMPEG_EXECUTABLE,
        '-loglevel', 'quiet',
        '-i', filepath,
        '-f', format_string,
        '-acodec', 'pcm_' + format_string,
        '-ar', str(samplerate),
        '-ac', str(channels),
        '-']
    print('ffmpeg load will start')
    # proc = asyncio.subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, startupinfo=STARTUPINFO)#, bufsize=64)
    async def runcmd(command):
        proc = await asyncio.create_subprocess_shell(command, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.DEVNULL, startupinfo=STARTUPINFO)#, bufsize=64)
        stdout, _ = await proc.communicate()

    # bytes_per_sample = numpy.dtype(in_type).itemsize
    # frame_size = bytes_per_sample * channels
    # chunk_size = frame_size * sr

    # print('ffmpeg load audio subprocess started')
    # with proc.stdout as stdout:
        # raw = stdout.read()
        raw = stdout
        # print(type(raw))
        audio = numpy.frombuffer(raw)#, dtype=in_type).astype(out_type)
        return audio

    audio = asyncio.run(runcmd(' '.join(command)))

    print('ffmpeg load ended')
    # p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, bufsize=4096, startupinfo=STARTUPINFO) # creationflags=subprocess.CREATE_NO_WINDOW,
    # bytes_per_sample = numpy.dtype(in_type).itemsize
    # frame_size = bytes_per_sample * channels
    # chunk_size = frame_size * sr # read in 1-second chunks
    # raw = b''
    # with p.stdout as stdout:
    #     while True:
    #         data = stdout.read(chunk_size)
    #         if data:
    #             raw += data
    #         else:
    #             break

    if channels > 1:
        audio = audio.reshape((-1, channels)).transpose()
    if audio.size == 0:
        return audio, samplerate
    if issubclass(out_type, numpy.floating):
        if normalize:
            peak = numpy.abs(audio).max()
            if peak > 0:
                audio /= peak
        elif issubclass(in_type, numpy.integer):
            audio /= numpy.iinfo(in_type).max
    print(len(audio))
    return audio


def ffms2_load_audio(filepath, samplerate=48000, mono=True, normalize=True, in_type=numpy.int16, out_type=numpy.float32):
    # asource = ffms2.AudioSource(filepath)
    # asource.init_buffer(asource.properties.NumSamples)
    # print(asource.properties.NumSamples)
    # print(4096*64000)
    # asource.init_buffer((4096*64000))
    # sound_np = asource.get_audio(0)
    # start = (4096*64)
    # while start < asource.properties.NumSamples:
    #     if start + (4096*64) > asource.properties.NumSamples:
    #         asource.init_buffer((4096*64) % ((start + (4096*64)) - asource.properties.NumSamples))

    #     sound_np = numpy.append(sound_np, asource.get_audio(start), axis=0)
    #     start += (4096*64)
    # # print(sound_np)
    # print(sound_np.shape)
    return []#sound_np


def ffmpeg_extract_subtitle(filepath, index):
    """Function to extract subtitle from video using ffmpeg"""
    command = [
        FFMPEG_EXECUTABLE,
        '-i',
        filepath,
        '-map',
        '0:' + str(index),
        os.path.join(path_tmp, 'subtitle.vtt')]
    subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, startupinfo=STARTUPINFO).wait()
    print('subtitle extracted')
    return os.path.join(path_tmp, 'subtitle.vtt')


def ffmpeg_load_metadata(filepath):
    """Function to read video's metadata"""
    command = [
        FFPROBE_EXECUTABLE,
        '-v',
        'quiet',
        '-print_format',
        'json',
        '-show_format',
        '-show_streams',
        filepath]
    proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, startupinfo=STARTUPINFO)
    json_file = False
    with proc.stdout as stdout:
        # test = stdout.read()
        # print(test)
        json_file = json.loads(stdout.read())

    return json_file


def generate_waveform_zoom(zoom, duration, waveform):
    """Function to calculate waveform from numpy"""
    positive_values = []
    negative_values = []
    parser = 0
    while parser < len(waveform):
        positive_values.append(numpy.amax(waveform[parser:parser+int(len(waveform) / (duration * zoom))]))
        negative_values.append(numpy.amin(waveform[parser:parser+int(len(waveform) / (duration * zoom))]))
        parser += int(len(waveform) / (duration * zoom))
    average = False
    parser = 0

    # bps = 120/60
    # chunk = int(44100/bps)
    # chunk = 4096

    return positive_values, negative_values, average


def generate_waveform_zoom2(zoom, duration, filepath):
    positive_values = []
    negative_values = []
    average = False

    asource = ffms2.AudioSource(filepath)

    parser = 0
    chunk = int(asource.properties.NumSamples / (duration * zoom))
    print(chunk)
    while parser < asource.properties.NumSamples:
        if parser + chunk > asource.properties.NumSamples:
            chunk = (asource.properties.NumSamples - parser)

        asource.init_buffer(chunk)
        sound_np = asource.get_audio(parser)

        if sound_np.ndim > 1:
            sound_np = numpy.mean(sound_np, axis=1)

        positive_values.append(numpy.amax(sound_np))
        negative_values.append(numpy.amin(sound_np))

        parser += chunk




    # counter = 0
    # chunk = asource.properties.SampleRate * 60
    # while counter < asource.properties.NumSamples:
    #     if counter + chunk > asource.properties.NumSamples:
    #         chunk = (asource.properties.NumSamples - counter)

    #     asource.init_buffer(chunk)
    #     sound_np = asource.get_audio(counter)

    #     if sound_np.ndim > 1:
    #         sound_np = numpy.mean(sound_np, axis=1)

    #     parser = 0
    #     while parser < len(sound_np):
    #         positive_values.append(numpy.amax(sound_np[parser:parser + int(len(sound_np) / (duration * zoom))]))
    #         negative_values.append(numpy.amin(sound_np[parser:parser + int(len(sound_np) / (duration * zoom))]))
    #         parser += int(asource.properties.NumSamples / (duration * zoom))

    #     counter += chunk

    return positive_values, negative_values, average


def return_waveform_zoom(self, qpixmap):
    """Function to return waveform's zoom"""
    self.video_metadata['waveform'][qpixmap[0]] = qpixmap[1]


class ThreadExtractWaveform2(QThread):
    """Thread to extract waveform"""
    command = Signal(list)
    filepath = ''
    duration = 60
    zoom = 100
    border_color = '#ff153450'
    fill_color = '#cc153450'

    class DrawPixmap(QPixmap):
        waveform_up = []
        waveform_down = []
        x_offset = 0
        waveformsize = .7
        border_color = '#ff153450'
        fill_color = '#cc153450'

        def paintEvent(self):
            painter = QPainter(self)
            painter.setRenderHint(QPainter.Antialiasing)

            painter.setPen(QPen(QColor(self.border_color), 1, Qt.SolidLine))
            painter.setBrush(QColor(self.fill_color))

            x_position = 0
            polygon = QPolygonF()
            for point in self.waveform_up:
                polygon.append(QPointF(x_position, (self.height() * .5) + (point * (self.waveformsize * 100))))
                x_position += 1
            for point in reversed(self.waveform_down):
                polygon.append(QPointF(x_position, (self.height() * .5) + (point * (self.waveformsize * 100))))
                x_position -= 1
            painter.drawPolygon(polygon)

            painter.end()

    def run(self):
        """Run function of thread to extract waveform"""
        if self.filepath:
            asource = ffms2.AudioSource(self.filepath)
            counter = 0
            chunk = 4096*100
            while counter < asource.properties.NumSamples:
                asource.init_buffer(chunk)
                sound_np = asource.get_audio(counter)

                positive_values = []
                negative_values = []
                parser = 0
                while parser < len(sound_np):
                    positive_values.append(numpy.amax(sound_np[parser:parser + int(len(sound_np) / (self.duration * self.zoom))]))
                    negative_values.append(numpy.amin(sound_np[parser:parser + int(len(sound_np) / (self.duration * self.zoom))]))
                    parser += int(len(sound_np) / (self.duration * self.zoom))


                # bps = 120/60
                # chunk = int(44100/bps)
                # chunk = 4096

                pixmap_width = len(positive_values)
                qpixmap = self.DrawPixmap(pixmap_width, 124)
                qpixmap.fill(QColor(0, 0, 0, 0))
                qpixmap.waveform_up = positive_values[parser:parser + pixmap_width]
                qpixmap.waveform_down = negative_values[parser:parser + pixmap_width]
                qpixmap.border_color = self.border_color
                qpixmap.fill_color = self.fill_color
                qpixmap.paintEvent()

                self.command.emit([self.zoom, qpixmap.toImage()])

                counter += chunk - (((asource.properties.NumSamples - (counter + chunk)) - chunk) if (counter + chunk) > asource.properties.NumSamples else 0)



#######################################################################
#
# Subtitld setup.py
#
#######################################################################

import os
import sys
import pydoc

from setuptools import setup, find_packages

import subtitld


def get_description(filename='README.md'):
    with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), filename), encoding='utf-8') as f:
        file = list(f)
    desc = ''
    for item in file[11: len(file)]:
        desc += item
    return desc


def get_data_files():
    files = []
    if sys.platform.startswith('linux') and 'QT_APPIMAGE' not in os.environ.keys():
        # appid = subtitld.__desktopid__
        files = [
            # ('share/icons/hicolor/16x16/apps', ['data/icons/hicolor/16x16/apps/{}.png'.format(appid)]),
            # ('share/icons/hicolor/16x16/apps', ['data/icons/hicolor/16x16/apps/{}.png'.format(appid)]),
            # ('share/icons/hicolor/22x22/apps', ['data/icons/hicolor/22x22/apps/{}.png'.format(appid)]),
            # ('share/icons/hicolor/24x24/apps', ['data/icons/hicolor/24x24/apps/{}.png'.format(appid)]),
            # ('share/icons/hicolor/32x32/apps', ['data/icons/hicolor/32x32/apps/{}.png'.format(appid)]),
            # ('share/icons/hicolor/48x48/apps', ['data/icons/hicolor/48x48/apps/{}.png'.format(appid)]),
            # ('share/icons/hicolor/64x64/apps', ['data/icons/hicolor/64x64/apps/{}.png'.format(appid)]),
            # ('share/icons/hicolor/128x128/apps', ['data/icons/hicolor/128x128/apps/{}.png'.format(appid)]),
            # ('share/icons/hicolor/256x256/apps', ['data/icons/hicolor/256x256/apps/{}.png'.format(appid)]),
            # ('share/icons/hicolor/512x512/apps', ['data/icons/hicolor/512x512/apps/{}.png'.format(appid)]),
            # ('share/icons/hicolor/scalable/apps', ['subtitld/graphics/subtitld.svg']),
            ('share/icons/hicolor/512x512/apps', ['subtitld/graphics/subtitld.png']),
            ('share/applications', ['subtitld.desktop']),
            # ('share/metainfo', ['data/appdata/{}.appdata.xml'.format(appid)]),
            # ('share/mime/packages', ['data/mime/{}.xml'.format(appid)]),
            # ('share/doc/subtitld', ['CHANGELOG', 'LICENSE', 'README.md'])
        ]

    return files


def get_package_files():
    files = {}

    files['subtitld'] = [
        'locale/*',
        'graphics/*',
    ]

    return files


def pip_notes():
    os.system('cls' if sys.platform == 'win32' else 'clear')
    pydoc.pager('''
    If installing via PyPi (Python Pip) on Linux then you need to know that subtitld
    depends on the following packages and distros. Install using your distro's
    software packager for a noticeably better integrated experience.

        ---[Ubuntu/Debian/Mint/etc]--------------------------

            python3-dev libmpv1 libmpv-dev python3-pyqt5
            python3-pyqt5.qtopengl python3-pyqt5.qtx11extras
            ffmpeg mediainfo python3-opengl

        ---[Arch Linux]--------------------------------------

            python mpv python-pyqt5 ffmpeg mediainfo

        ---[Fedora]------------------------------------------

            python3-devel mpv-libs mpv-libs-devel python3-qt5
            ffmpeg mediainfo python3-pyopengl

        ---[openSUSE]----------------------------------------

            python3-devel libmpv1 mpv-devel python3-qt5
            ffmpeg mediainfo

    You need to build a Python extension module before you can run the
    app directly from source code. This is done for you automatically by
    the package installers but if you wish to simply run the app direct
    from source without having to install it (i.e. python3 setup.py install)
    you can do so by building the extension module with the following
    setuptools command, run from the top-most extracted source code folder:

        $ python3 setup.py build_ext -i

    And to then run the app directly from source, from the same top-most
    source code folder:

        $ python3 -m subtitld

    To view all console output for debugging or general interest then
    append the debug parameter:

        $ python3 -m subtitld --debug

    Make sure you build the extension module AFTER installing the
    dependencies covered above, in particular libmpv and the mpv + python3
    dev headers are all needed for it to compile successfully. Compiled
    extension modules under subtitld/libs will look like:

        mpv.cpython-36m-x86_64-linux-gnu.so [linux]
        mpv.cp36-win_amd64.pyd              [win32]


    Get more information on:

    https://subtitld.org
''')


# --------------------------------------------------------------------------- #

setup_requires = ['setuptools']
install_requires = [
    'PySide6',
    'pyopengl',
    'python-mpv==0.5.2',
    'ffms2',
    'numpy',
    'cffi',
    'requests',
    'pycaption',
    # 'captionstransformer @ git+ssh://git@github.com/toutpt/captionstransformer',
    'pysubs2',
    'lxml',
    'clean-text[gpl]',
    'html5lib==1.0b8',
    # 'scenedetect',
    # 'opencv-python',
    # 'autosub3',
    'google-trans-new',
    'SpeechRecognition',
    'beautifulsoup4<4.10,>=4.8.1',
    'python-docx',
    'chardet',
    'google-api-python-client',
    'pysrt',
    'certifi'
]

# --------------------------------------------------------------------------- #

try:
    # begin setuptools installer
    result = setup(
        app=['subtitld/__main__.py'],
        name=subtitld.__appname__.lower(),
        version=subtitld.__version__,
        author=subtitld.__author__,
        author_email=subtitld.__email__,
        description='Subtitld',
        long_description=get_description(),
        url=subtitld.__website__,
        license='GPL3',
        packages=find_packages(include=['subtitld', 'subtitld.*']),
        setup_requires=setup_requires,
        install_requires=install_requires,
        data_files=get_data_files(),
        package_data=get_package_files(),
        include_package_data=True,
        entry_points={'gui_scripts': ['subtitld = subtitld.__main__:main']},
        keywords='subtitld',
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Environment :: X11 Applications :: Qt',
            'Intended Audience :: End Users/Desktop',
            'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
            'Natural Language :: English',
            'Operating System :: POSIX',
            'Topic :: Multimedia',
            'Programming Language :: Python :: 3 :: Only'
        ],
        options={'py2app': {
            'argv_emulation': True,
            # 'iconfile': 'src/Icon.icns',  # optional
            # 'plist': 'src/Info.plist',    # optional
        }},
    )
except BaseException:
    if subtitld.__ispypi__:
        pip_notes()
    raise



"""Module for subtitle list panel

"""
from PySide6.QtWidgets import QFrame, QTextEdit, QVBoxLayout, QPushButton
from PySide6.QtGui import QTextCursor

from subtitld.interface import subtitles_panel
from subtitld.modules import utils


def add_widgets(self):
    self.subtitles_panel_markdown_widget = QFrame()
    self.subtitles_panel_markdown_widget.setObjectName('subtitles_panel_markdown_widget')
    self.subtitles_panel_markdown_widget.setLayout(QVBoxLayout())
    self.subtitles_panel_markdown_widget.layout().setContentsMargins(20, 0, 0, 0)
    self.subtitles_panel_markdown_widget.layout().setSpacing(0)

    class subtitles_panel_markdown_qtextedit(QTextEdit):
        def focusOutEvent(widget, event):
            update_subtitles_panel_markdown(self)
            event.accept()

    self.subtitles_panel_markdown_qtextedit = subtitles_panel_markdown_qtextedit()
    self.subtitles_panel_markdown_qtextedit.setObjectName('subtitles_panel_markdown_qtextedit')
    # self.subtitles_panel_markdown_qtextedit.focusInEvent.connect(lambda: subtitles_panel_markdown_qtextedit_cursorpositionchanged(self))
    # self.subtitles_panel_markdown_qtextedit.focusOutEvent.connect(lambda: update_subtitles_panel_markdown(self))
    self.subtitles_panel_markdown_qtextedit.cursorPositionChanged.connect(lambda: subtitles_panel_markdown_qtextedit_cursorpositionchanged(self))
    self.subtitles_panel_markdown_qtextedit.textChanged.connect(lambda: subtitles_panel_markdown_qtextedit_textchanged(self))
    self.subtitles_panel_markdown_widget.layout().addWidget(self.subtitles_panel_markdown_qtextedit)

    self.subtitles_panel_stackedwidgets.addWidget(self.subtitles_panel_markdown_widget)


def add_button(self):
    self.subtitles_panel_widget_button_markdown = QPushButton()
    self.subtitles_panel_widget_button_markdown.setObjectName('subtitles_panel_widget_button_markdown')
    self.subtitles_panel_widget_button_markdown.setProperty('class', 'subtitles_panel_left_button')
    self.subtitles_panel_widget_button_markdown.setCheckable(True)
    self.subtitles_panel_widget_button_markdown.setFixedWidth(23)
    self.subtitles_panel_widget_button_markdown.clicked.connect(lambda vision: subtitles_panel.update_subtitles_panel_widget_vision(self, 'markdown'))
    self.subtitles_panel_widget_buttons_vbox.addWidget(self.subtitles_panel_widget_button_markdown)


def subtitles_panel_markdown_qtextedit_cursorpositionchanged(self):
    # print('blah')
    position = self.subtitles_panel_markdown_qtextedit.textCursor().position()
    # print(position)
    cursor = 0
    # # markdown_text = ''
    for subtitle in sorted(self.subtitles_list):
        cursor += len(str("{:.3f}".format(subtitle[0])))
        next_index = self.subtitles_list.index(subtitle) + 1
        if not next_index >= len(self.subtitles_list) and not self.subtitles_list[next_index][0] - 0.001 == subtitle[0] + subtitle[1]:
            cursor += len(' - ' + str("{:.3f}".format(subtitle[0] + subtitle[1])))
        cursor += len('\n')

        cursor += len(str(subtitle[2]) + '\n\n')
        if cursor > position:
            self.selected_subtitle = subtitle
            break

    if self.selected_subtitle:
        if not self.selected_subtitle[0] < self.player_widget.position < self.selected_subtitle[0] + self.selected_subtitle[1]:
            self.player_widget.seek(self.selected_subtitle[0] + (self.selected_subtitle[1] * .5))

            self.timeline.update_scrollbar(self, position='middle')


def subtitles_panel_markdown_qtextedit_textchanged(self):
    subtitles_panel_markdown_qtextedit_update_subtitles_list(self)


def subtitles_panel_markdown_qtextedit_update_subtitles_list(self):
    sub_list = []

    last_time = []
    last_text = ''
    s = 0
    for line in self.subtitles_panel_markdown_qtextedit.toPlainText().split('\n'):
        if all(utils.is_float(t) for t in line.strip().replace(' - ', ' ').split(' ')):
            this_time = [
                float(line.strip().split(' - ')[0])
            ]

            if ' - ' in line.strip():
                this_time.append(float(line.strip().split(' - ')[-1]))
            else:
                this_time.append(0)

            if last_time:
                sub_list.append(
                    [
                        float(last_time[0]),
                        (float(last_time[1]) - float(last_time[0])) if last_time[1] else (float(this_time[0]) - .001 - float(last_time[0])),
                        last_text.strip()
                    ]
                )
                last_text = ''
                s += 1

            last_time = this_time

        else:
            last_text += line + '\n'

    if last_time:
        sub_list.append(
            [
                float(last_time[0]),
                (float(last_time[1]) - float(last_time[0])) if last_time[1] else (float(this_time[0]) - .001 - float(last_time[0])),
                last_text.strip()
            ]
        )
        last_text = ''

    self.selected_subtitle = False

    # Sanitize subtitles so there is no overlaping subtitles?

    self.subtitles_list = sorted(sub_list)

    self.timeline.update(self)


def update_subtitles_panel_markdown(self, selection=False):
    cursor = self.subtitles_panel_markdown_qtextedit.textCursor()
    # old_scrollbar_value = self.subtitles_panel_markdown_qtextedit.verticalScrollBar().value()

    selected_subtitle_found = False
    position = 0
    markdown_text = ''
    for subtitle in sorted(self.subtitles_list):
        markdown_text += '<small><b>' + str("{:.3f}".format(subtitle[0]))
        if not selected_subtitle_found:
            position += len(str("{:.3f}".format(subtitle[0])))

        next_index = self.subtitles_list.index(subtitle) + 1
        if next_index < len(self.subtitles_list) and not self.subtitles_list[next_index][0] - 0.001 == subtitle[0] + subtitle[1] or (next_index == len(self.subtitles_list) and not self.video_metadata['duration'] - 0.001 == subtitle[0] + subtitle[1]):
            markdown_text += ' - ' + str("{:.3f}".format(subtitle[0] + subtitle[1]))
            if not selected_subtitle_found:
                position += len(' - ' + str("{:.3f}".format(subtitle[0] + subtitle[1])))

        markdown_text += '</b></small><br/>'
        if not selected_subtitle_found:
            position += len('\n')

        markdown_text += str(subtitle[2]) + '<br/><br/>'

        if self.selected_subtitle == subtitle:
            selected_subtitle_found = True
            # position -= 2

        if not selected_subtitle_found:
            position += len(str(subtitle[2]) + '\n\n')

    self.subtitles_panel_markdown_qtextedit.blockSignals(True)

    self.subtitles_panel_markdown_qtextedit.setHtml(markdown_text)

    if selection:
        cursor.setPosition(position + selection[0])
        cursor.setPosition(position + selection[0] + selection[1], QTextCursor.KeepAnchor)
    elif position < len(markdown_text):
        cursor.setPosition(position)

    self.subtitles_panel_markdown_qtextedit.setTextCursor(cursor)
    # self.subtitles_panel_markdown_qtextedit.verticalScrollBar().setValue(old_scrollbar_value)

    # self.subtitles_panel_markdown_qtextedit.setFocus(Qt.TabFocusReason)

    self.subtitles_panel_markdown_qtextedit.blockSignals(False)
